# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.flax_layers."""

import functools
import itertools
from typing import Any, Dict, Mapping
from unittest import mock

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import flax_layers
from aqt.jax_legacy.jax import fp_cast
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import primitives
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import quantization
from aqt.jax_legacy.jax import shape_utils
from aqt.jax_legacy.jax import test_utils
from aqt.jax_legacy.jax.quantization import QuantOps
from aqt.jax_legacy.jax.quantization import QuantType
import flax
from flax import linen as nn
from flax import traverse_util
import jax
from jax import config
from jax import dtypes
from jax import lax
from jax import random
from jax.nn import initializers
import jax.numpy as jnp
import numpy as onp


FLAGS = flags.FLAGS

# fp-1-4-3
#    1: sign
#    4: number of exponent-bits, (bias = 11), range: -11, ..., 4
#    3: number of significand-bits (excluding hidden-bit)
fp143_scaled = QuantOps.FloatQuant(
    is_scaled=True,
    fp_spec=QuantOps.FloatQuant.FloatPrec(
        exp_min=-11,
        exp_max=4,
        sig_bits=3,
    ),
)
fp143_unscaled = QuantOps.FloatQuant(
    is_scaled=False,
    fp_spec=QuantOps.FloatQuant.FloatPrec(
        exp_min=-11,
        exp_max=4,
        sig_bits=3,
    ),
)

METADATA_KEY = '__save_format_metadata__'


def filter_out_metadata(params: Mapping[str, Any]) -> Dict[str, Any]:
  """Removes "__save_format_metadata__" entries from a parameter tree."""
  result = {}
  for k, v in params.items():
    if k == METADATA_KEY:
      continue
    if isinstance(v, Mapping):
      v = filter_out_metadata(v)
      if not v:
        continue
    result[k] = v
  return result


def param_dtypes_shapes_axes(params: Mapping[str, Any],
                             params_axes: Mapping[str, Any]) -> Dict[str, Any]:
  """Construct a tree of param info including dtypes, shapes, and axis names.

  The leaf of the constructed dtree are of format [<dtype>, <axis_dim>, ...],
  where each <axis_dim> is of format <axis_name>=<dim>.

  Args:
    params: Model params.
    params_axes: Axis annotations, typically under state["params_axes"].

  Returns:
    A pytree with params info.
  """
  params = filter_out_metadata(params)
  params_axes = filter_out_metadata(params_axes)
  params = flax.core.unfreeze(params)  # pytype: disable=wrong-arg-types

  def remove_axes_suffix(ks):
    if not ks[-1].endswith('_axes'):
      raise ValueError(
          f'Param axes name should end with `_axes`, found {ks[-1]}')
    return tuple(ks[:-1]) + (ks[-1][:-len('_axes')],)

  params_axes = flax.core.unfreeze(params_axes)  # pytype: disable=wrong-arg-types
  flatten_axes = {
      remove_axes_suffix(ks): v
      for ks, v in traverse_util.flatten_dict(params_axes).items()
  }
  params_axes = traverse_util.unflatten_dict(flatten_axes)

  def _create_entry(param, param_axes):
    output = [str(param.dtype)]
    # The param axes should be paired with param dimension, so we check that.
    if param.ndim != len(param_axes.names):
      raise ValueError('Length of param dimension does not match axes, '
                       f'{param.shape} != {param_axes.names}.')
    for dim, axis_name in zip(param.shape, param_axes.names):
      output.append(f'{axis_name}={dim}')
    return output

  return jax.tree_map(_create_entry, params, params_axes)


class ConvAqtTest(parameterized.TestCase):
  """Tests for ConvAqt layer."""

  def setUp(self):
    super(ConvAqtTest, self).setUp()
    test_utils.configure_jax()
    quantization.DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING = True
    self.rng_key = random.PRNGKey(0)

  def tearDown(self):
    config.update('jax_numpy_rank_promotion', 'warn')
    super(ConvAqtTest, self).tearDown()

  def init_model_with_1_layer(self,
                              inputs,
                              num_features,
                              kernel_size,
                              kernel_init=flax_layers.default_kernel_init,
                              weight_prec=None,
                              quant_act=None,
                              weight_half_shift=False):
    """Create and initialize a flax model with a single ConvAqt layer."""
    layer_kwargs = {
        'kernel_init': kernel_init,
        'features': num_features,
        'use_bias': False,
        'dynamic_context': quant_config.DynamicContext(update_bounds=False),
        'paxis_name': 'batch',
        'train': False,
        'kernel_size': kernel_size,
        'dtype': jnp.float32
    }
    layer_class = flax_layers.ConvAqt
    layer_kwargs['hparams'] = flax_layers.ConvAqt.HParams(
        weight_prec=weight_prec,
        quant_act=quant_act,
        quant_type=QuantType.FAKE_QUANT,
        weight_half_shift=weight_half_shift,
    )
    conv_module = layer_class(**layer_kwargs)
    initial_state = conv_module.init(self.rng_key, jnp.zeros(inputs.shape))
    return conv_module, initial_state

  # Following ConvAqt tests adapted from
  # Flax Conv tests.
  @parameterized.named_parameters(
      dict(testcase_name='float', weight_prec=None),
      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
      dict(testcase_name='quant_2bit', weight_prec=2),
  )
  def test_conv(self, weight_prec=None):
    x = jnp.ones((1, 8, 8, 3))
    conv_module = flax_layers.ConvAqt(
        features=4,
        kernel_size=(3, 3),
        padding='VALID',
        paxis_name='batch',
        dynamic_context=quant_config.DynamicContext(update_bounds=False),
        train=False,
        hparams=flax_layers.ConvAqt.HParams(
            weight_prec=weight_prec,
            quant_act=None,
            quant_type=QuantType.FAKE_QUANT,
            weight_half_shift=False),
        kernel_init=initializers.ones,
        bias_init=initializers.ones,
        dtype=jnp.float32)

    y, state = conv_module.init_with_output(self.rng_key, x)
    self.assertEqual(state['params']['kernel'].shape, (3, 3, 3, 4))
    test_utils.assert_all_close_prec(y, onp.full((1, 6, 6, 4), 28.),
                                     weight_prec)

  @parameterized.named_parameters(
      dict(testcase_name='float', weight_prec=None),
      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
      dict(testcase_name='quant_2bit', weight_prec=2),
  )
  def test_group_conv(self, weight_prec=None):
    x = jnp.ones((1, 8, 8, 4))
    conv_module = flax_layers.ConvAqt(
        features=4,
        kernel_size=(3, 3),
        feature_group_count=2,
        padding='VALID',
        paxis_name='batch',
        dynamic_context=quant_config.DynamicContext(update_bounds=False),
        train=False,
        hparams=flax_layers.ConvAqt.HParams(
            weight_prec=weight_prec,
            quant_act=None,
            quant_type=QuantType.FAKE_QUANT,
            weight_half_shift=False),
        kernel_init=initializers.ones,
        bias_init=initializers.ones,
        dtype=jnp.float32)
    y, state = conv_module.init_with_output(self.rng_key, x)
    self.assertEqual(state['params']['kernel'].shape, (3, 3, 2, 4))
    test_utils.assert_all_close_prec(y, onp.full((1, 6, 6, 4), 19.),
                                     weight_prec)

  @parameterized.named_parameters(
      dict(testcase_name='conv_quant_8bit', weight_prec=8),
      dict(testcase_name='conv_quant_4bit', weight_prec=4),
      dict(testcase_name='conv_quant_2bit', weight_prec=2),
  )
  def test_full_range_integer_weights_should_give_precise_output(
      self, weight_prec):
    # If weights are ints (already quantized) and
    # max(abs(weights[:, ch])) == 2**(prec-1)-1 in each channel,
    # no quantization error should be introduced.

    num_features = 256
    input_dim = 3
    inputs = random.uniform(self.rng_key, shape=(1, 16, 16, input_dim))
    kernel_size = (3, 3)
    model, state = self.init_model_with_1_layer(
        inputs, num_features, kernel_size, weight_prec=weight_prec)
    minval = -2**(weight_prec - 1) + 1
    maxval = 2**(weight_prec - 1) - 1

    full_range_integer_weights = random.randint(
        self.rng_key, kernel_size + (input_dim, num_features), minval,
        maxval + 1)

    # manually set one value in each output dim of weights to be exactly maxval
    full_range_integer_weights = (
        full_range_integer_weights.at[0, 0, :].set(maxval))
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = full_range_integer_weights
    state = flax.core.freeze(state)
    outputs = model.apply(state, inputs)

    dimension_numbers = nn.linear._conv_dimension_numbers(inputs.shape)  # pylint: disable=protected-access
    exp_outputs = lax.conv_general_dilated(
        inputs,
        jnp.asarray(state['params']['kernel'], jnp.float32), (1, 1),
        'SAME',
        lhs_dilation=None,
        rhs_dilation=None,
        dimension_numbers=dimension_numbers,
        feature_group_count=1,
        precision=jax.lax.Precision.DEFAULT)

    onp.testing.assert_array_equal(outputs, exp_outputs)

  @parameterized.named_parameters(
      dict(testcase_name='conv_quant_4bit', weight_prec=4),
      dict(testcase_name='conv_quant_2bit', weight_prec=2),
  )
  def test_full_range_integer_weights_with_float_scale_should_give_close_output(
      self, weight_prec):
    # If weights are ints (already quantized) with
    # max(abs(weights[..., ch])) == 2**(prec-1)-1 in each channel
    # and if these integer weights are multiplied by a float scale,
    # the resulting error should still be very small (just float rounding).

    num_features = 256
    input_dim = 3
    inputs = random.uniform(self.rng_key, shape=(1, 16, 16, input_dim))
    kernel_size = (3, 3)
    model, state = self.init_model_with_1_layer(
        inputs, num_features, kernel_size, weight_prec=weight_prec)
    minval = -2**(weight_prec - 1) + 1
    maxval = 2**(weight_prec - 1) - 1

    full_range_integer_weights = random.randint(
        self.rng_key, kernel_size + (input_dim, num_features), minval,
        maxval + 1)
    # manually set one value in each output dim of weights to be exactly maxval
    full_range_integer_weights = (
        full_range_integer_weights.at[0, 0, :].set(maxval))

    # (batch_size, spatial_dim, spatial_dim, num_features)
    float_scale = jax.random.uniform(self.rng_key, (1, 1, 1, num_features))
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = full_range_integer_weights * float_scale
    state = flax.core.freeze(state)
    outputs = model.apply(state, inputs)
    dimension_numbers = nn.linear._conv_dimension_numbers(inputs.shape)  # pylint: disable=protected-access
    exp_outputs = lax.conv_general_dilated(
        inputs,
        jnp.asarray(state['params']['kernel'], jnp.float32), (1, 1),
        'SAME',
        lhs_dilation=None,
        rhs_dilation=None,
        dimension_numbers=dimension_numbers,
        feature_group_count=1,
        precision=jax.lax.Precision.DEFAULT)
    # We know that the noise should be proportional to the square root of
    # input_dim and inversely proportional to 2**weight_prec.
    # The following tol_const was obtained experimentally and should be derived
    # more systematically.
    tol_const = 5e-02
    onp.testing.assert_allclose(
        outputs,
        exp_outputs,
        rtol=jnp.sqrt(input_dim) * 2**(-weight_prec) * tol_const)

  @parameterized.named_parameters(
      dict(
          testcase_name='conv_quant_8bit',
          weight_prec=8,
          weight_scale=onp.array([1, 2, 4, 8])),
      dict(
          testcase_name='conv_quant_4bit',
          weight_prec=4,
          weight_scale=onp.array([1, 2, 4, 8])),
      dict(
          testcase_name='conv_quant_2bit',
          weight_prec=2,
          weight_scale=onp.array([1, 2, 4, 8])),
  )
  def test_weight_invariance_to_power_of_2_weight_scaling(
      self, weight_prec, weight_scale):
    # Scaling the weights before quantization by a power of 2 per channel should
    # also scale the output exactly by the same scale.

    num_features = 4
    assert num_features == weight_scale.shape[-1]
    input_dim = 3
    inputs = random.uniform(self.rng_key, shape=(1, 16, 16, input_dim))
    kernel_size = (3, 3)
    model, state = self.init_model_with_1_layer(
        inputs,
        4,
        kernel_size,
        weight_prec=weight_prec,
        weight_half_shift=False)

    weights = random.uniform(
        self.rng_key, shape=kernel_size + (input_dim, num_features))
    weight_scale = weight_scale[jnp.newaxis, jnp.newaxis, jnp.newaxis, :]
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = weights
    outputs_without_scaling = model.apply(flax.core.freeze(state), inputs)
    state['params']['kernel'] = jnp.multiply(weights, weight_scale)
    outputs_with_scaling = model.apply(flax.core.freeze(state), inputs)

    onp.testing.assert_array_equal(outputs_without_scaling * weight_scale,
                                   outputs_with_scaling)

  def test_1_bit_makes_all_weight_equal_to_zero(self):
    num_features = 4
    input_dim = 3
    inputs = random.uniform(self.rng_key, shape=(1, 32, 32, input_dim))
    kernel_size = (3, 3)
    model, state = self.init_model_with_1_layer(
        inputs, num_features, kernel_size, weight_prec=1)
    weights = random.uniform(
        self.rng_key, shape=kernel_size + (input_dim, num_features))
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = weights
    outputs = model.apply(flax.core.freeze(state), inputs)
    onp.testing.assert_array_equal(outputs, onp.zeros(
        (1, 32, 32, num_features)))

  @parameterized.named_parameters(
      dict(
          testcase_name='conv_quant_8bit',
          weight_prec=8,
          acts_prec=None,
          fixed_bounds=True),
      dict(
          testcase_name='conv_quant_4bit',
          weight_prec=4,
          acts_prec=None,
          fixed_bounds=True),
      dict(
          testcase_name='conv_quant_2bit',
          weight_prec=2,
          acts_prec=None,
          fixed_bounds=True),
      dict(
          testcase_name='conv_signed_input_quant_8bit',
          weight_prec=None,
          acts_prec=8,
          fixed_bounds=True),
      dict(
          testcase_name='conv_signed_input_quant_4bit',
          weight_prec=None,
          acts_prec=4,
          fixed_bounds=True),
      dict(
          testcase_name='conv_signed_input_auto_quant_8bit',
          weight_prec=None,
          acts_prec=8,
          fixed_bounds=False),
      dict(
          testcase_name='conv_signed_input_auto_quant_4bit',
          weight_prec=None,
          acts_prec=4,
          fixed_bounds=False),
      dict(
          testcase_name='conv_signed_input_quant_2bit',
          weight_prec=None,
          acts_prec=2,
          fixed_bounds=True),
  )
  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_quantized_weights_and_symmetrics_acts_should_call_clip_and_round(
      self, floor_with_gradient, round_with_gradient, weight_prec, acts_prec,
      fixed_bounds):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x

    if fixed_bounds:
      bounds = 6.0
    else:
      bounds = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        prec=acts_prec,
        bounds=bounds,
        half_shift=False)
    num_features = 4
    input_dim = 3
    inputs = jnp.ones((1, 32, 32, input_dim), dtype=jnp.float32)
    kernel_size = (3, 3)
    model, state = self.init_model_with_1_layer(
        inputs,
        num_features,
        kernel_size,
        weight_prec=weight_prec,
        quant_act=quant_act,
        weight_half_shift=False)

    round_with_gradient.reset_mock()
    floor_with_gradient.reset_mock()

    outputs = model.apply(state, inputs)

    self.assertEqual(
        outputs.shape,
        (inputs.shape[0], inputs.shape[1], inputs.shape[2], num_features))
    round_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(round_with_gradient.call_count, 1)
    floor_with_gradient.assert_not_called()

  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_without_quantized_weights_should_not_call_quantization_ops(
      self, floor_with_gradient, round_with_gradient):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x
    inputs = jnp.ones((1, 32, 32, 3), dtype=jnp.float32)
    model, state = self.init_model_with_1_layer(inputs, 4, (3, 3))
    _ = model.apply(state, inputs)
    round_with_gradient.assert_not_called()
    floor_with_gradient.assert_not_called()


class DenseAqtTest(parameterized.TestCase):
  """Tests for DenseAqt layer."""

  def setUp(self):
    super(DenseAqtTest, self).setUp()
    test_utils.configure_jax()
    quantization.DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING = True
    self.rng_key = random.PRNGKey(0)

  def tearDown(self):
    config.update('jax_numpy_rank_promotion', 'warn')
    super(DenseAqtTest, self).tearDown()

  def init_model_with_1_layer(self,
                              inputs,
                              num_features,
                              kernel_init=flax_layers.default_kernel_init,
                              weight_prec=None,
                              quant_act=None,
                              weight_half_shift=False,
                              kernel_axis_names=None,
                              train=False,
                              possibly_use_quantized_vars=False,
                              quant_type=QuantType.FAKE_QUANT):
    """Create and initialize a flax model with a single DenseAqt layer."""
    dynamic_context = quant_config.DynamicContext(
        update_bounds=False, collect_acts_stats=False)
    layer_kwargs = {
        'kernel_init': kernel_init,
        'features': num_features,
        'use_bias': False,
        'dynamic_context': dynamic_context,
        'paxis_name': 'batch',
        'train': train,
        'dtype': jnp.float32,
        'kernel_axis_names': kernel_axis_names,
        'possibly_use_quantized_vars': possibly_use_quantized_vars
    }
    layer_kwargs['hparams'] = flax_layers.DenseAqt.HParams(
        weight_prec=weight_prec,
        quant_act=quant_act,
        quant_type=quant_type,
        weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
        weight_half_shift=weight_half_shift)

    dense_module = flax_layers.DenseAqt(**layer_kwargs)
    initial_state = dense_module.init(
        self.rng_key, jnp.zeros(inputs.shape), padding_mask=None)
    return dense_module, initial_state

  def test_padding(self):
    """Test that padding results in the right statistics being collected."""
    # Exact values don't matter here, we just need code to think it's using
    # dynamic bounds so it gathers activation statistics
    bounds = get_bounds.GetBounds.Hyper(
        initial_bound=0.0,
        stddev_coeff=1.0,
        absdev_coeff=0.0,
        mix_coeff=1.0,
        reset_stats=False,
        granularity=quant_config.QuantGranularity.PER_CHANNEL)
    quant_act = flax_layers.QuantOps.ActHParams(
        input_distribution=flax_layers.QuantOps.ActHParams.InputDistribution
        .SYMMETRIC,
        prec=8,
        bounds=bounds,
        half_shift=False)
    hparams = flax_layers.DenseAqt.HParams(
        quant_type=flax_layers.QuantType.FAKE_QUANT,
        weight_prec=8,
        quant_act=quant_act,
        weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
        weight_half_shift=False)
    module = flax_layers.DenseAqt(
        hparams=hparams,
        features=1,
        paxis_name=None,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=True, collect_acts_stats=False),
        train=True,
        dtype=jnp.float32)

    # Simulate an input with a batch size of 2, three tokens per example, two
    # channels per token
    x = jnp.arange(12).astype(jnp.float32).reshape((2, 3, 2))
    # Reshape it to have dimensions [batch, feature]
    x = x.reshape(6, 2)

    initial_state = module.init(self.rng_key, x, padding_mask=None)

    # Check that the per-channel activation statistics are as expected with no
    # padding
    _, state_nopadding = module.apply(
        initial_state, x, padding_mask=None, mutable='get_bounds')
    expected_means = onp.array([[(0 + 2 + 4 + 6 + 8 + 10) / 6,
                                 (1 + 3 + 5 + 7 + 9 + 11) / 6]])
    actual_means = state_nopadding['get_bounds']['GetBounds_0']['stats'].mean
    onp.testing.assert_allclose(actual_means, expected_means)

    # Now we pad out some of the tokens (chosen arbitrarily) and check that the
    # computed per-channel stats are the means of the non-padding tokens only
    # Exclude the second and third tokens from the first batch and the first
    # token from the second batch.
    padding_mask = jnp.array([[True, False, False], [False, True, True]])
    # Reshape it to have dimensions [batch, feature]
    padding_mask = padding_mask.reshape(6, 1)
    _, state_padding = module.apply(
        initial_state, x, padding_mask=padding_mask, mutable='get_bounds')
    expected_means = onp.array([[(0 + 8 + 10) / 3, (1 + 9 + 11) / 3]])
    actual_means = state_padding['get_bounds']['GetBounds_0']['stats'].mean
    onp.testing.assert_allclose(actual_means, expected_means)

  @parameterized.named_parameters(
      dict(testcase_name='dense_float', weight_prec=None),
      dict(
          testcase_name='dense_quant_fp143_scaled',
          weight_prec=fp143_scaled,
      ),
      dict(
          testcase_name='dense_quant_fp143',
          weight_prec=fp143_unscaled,
      ),
      dict(testcase_name='dense_quant_8bit', weight_prec=8),
      dict(testcase_name='dense_quant_4bit', weight_prec=4),
      dict(testcase_name='dense_quant_2bit', weight_prec=2),
  )
  def test_ones_weights_should_give_precise_output(self, weight_prec):
    """If all weights are 1, no quantization error should be introduced."""
    inputs = random.uniform(self.rng_key, shape=(2, 3))
    model, state = self.init_model_with_1_layer(
        inputs,
        num_features=4,
        kernel_init=initializers.ones,
        weight_prec=weight_prec,
        weight_half_shift=False)
    outputs = model.apply(state, inputs, padding_mask=None)
    exp_outputs = jnp.matmul(inputs, state['params']['kernel'])
    onp.testing.assert_array_equal(outputs, exp_outputs)

  def test_logical_axis_names(self):
    inputs = random.uniform(self.rng_key, shape=(2, 3))
    _, state = self.init_model_with_1_layer(
        inputs,
        num_features=4,
        kernel_init=initializers.ones,
        weight_prec=8,
        weight_half_shift=False,
        kernel_axis_names=('embed', 'mlp'))

    self.assertDictEqual(
        param_dtypes_shapes_axes(state['params'], state['params_axes']),
        {'kernel': ['float32', 'embed=3', 'mlp=4']})

  @parameterized.named_parameters(
      dict(testcase_name='dense_quant_8bit', weight_prec=8),
      dict(testcase_name='dense_quant_4bit', weight_prec=4),
      dict(testcase_name='dense_quant_2bit', weight_prec=2),
  )
  def test_full_range_integer_weights_should_give_precise_output(
      self, weight_prec):
    # If weights are ints (already quantized) and
    # max(abs(weights[:, ch])) == 2**(prec-1)-1 in each channel,
    # no quantization error should be introduced.
    num_features = 256
    input_dim = 1024
    inputs = random.uniform(self.rng_key, shape=(2, input_dim))
    model, state = self.init_model_with_1_layer(
        inputs, num_features, weight_prec=weight_prec, weight_half_shift=False)
    minval = -2**(weight_prec - 1) + 1
    maxval = 2**(weight_prec - 1) - 1

    full_range_integer_weights = random.randint(self.rng_key,
                                                (input_dim, num_features),
                                                minval, maxval + 1)

    # manually set one value in each output dim of weights to be exactly maxval
    full_range_integer_weights = full_range_integer_weights.at[0, :].set(maxval)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = full_range_integer_weights
    state = flax.core.freeze(state)
    outputs = model.apply(state, inputs, padding_mask=None)
    exp_outputs = jnp.matmul(inputs, state['params']['kernel'])
    onp.testing.assert_array_equal(outputs, exp_outputs)

  @parameterized.named_parameters(
      # TODO(shivaniagrawal): this test is flaky and fails with rtol=0.0004
      # with given rtol=0.0001
      # dict(
      #     testcase_name='dense_quant_8bit',
      #     layer_class=flax_layers.DenseAqt,
      #     weight_prec=8),
      dict(testcase_name='dense_quant_4bit', weight_prec=4),
      dict(testcase_name='dense_quant_2bit', weight_prec=2),
  )
  def test_full_range_integer_weights_with_float_scale_should_give_close_output(
      self, weight_prec):
    # If weights are ints (already quantized) with
    # max(abs(weights[:, ch])) == 2**(prec-1)-1 in each channel
    # and if these integer weights are multiplied by a float scale,
    # the resulting error should still be very small (just float rounding).

    num_features = 256
    input_dim = 1024
    inputs = random.uniform(self.rng_key, shape=(2, input_dim))
    model, state = self.init_model_with_1_layer(
        inputs, num_features, weight_prec=weight_prec)
    minval = -2**(weight_prec - 1) + 1
    maxval = 2**(weight_prec - 1) - 1

    full_range_integer_weights = random.randint(self.rng_key,
                                                (input_dim, num_features),
                                                minval, maxval + 1)

    # manually set one value in each output dim of weights to be exactly maxval
    full_range_integer_weights = full_range_integer_weights.at[0, :].set(maxval)

    float_scale = jax.random.uniform(self.rng_key, (1, num_features))
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = full_range_integer_weights * float_scale
    state = flax.core.freeze(state)
    outputs = model.apply(state, inputs, padding_mask=None)
    exp_outputs = jnp.matmul(inputs, state['params']['kernel'])
    # TODO(wanglisa): Determine how much noise is expected for following test.
    # We know that the noise should be proportional to the square root of
    # input_dim and inversely proportional to 2**weight_prec.
    # The following tol_const was obtained experimentally and should be derived
    # more systematically.
    tol_const = 8e-04
    onp.testing.assert_allclose(
        outputs,
        exp_outputs,
        rtol=jnp.sqrt(input_dim) * 2**(-weight_prec) * tol_const)

  @parameterized.named_parameters(
      # dict(
      #     testcase_name='dense_quant_8bit',
      #     weight_prec=8),
      # TODO(shivaniagrawal): fix the above test, test above doesn't follow
      # the expected tolerance. Expected absolute difference = 0.188386,
      # actual absolute difference: 0.20296225
      dict(
          testcase_name='dense_quant_fp143_scaled',
          weight_prec=fp143_scaled,
      ),
      dict(testcase_name='dense_quant_fp143', weight_prec=fp143_unscaled),
      dict(testcase_name='dense_quant_4bit', weight_prec=4),
      dict(testcase_name='dense_quant_2bit', weight_prec=2),
  )
  def test_float_weights_should_give_close_output(self, weight_prec):
    inputs = random.uniform(self.rng_key, shape=(2, 3))
    model, state = self.init_model_with_1_layer(
        inputs, num_features=4, weight_prec=weight_prec)
    float_weights = jnp.linspace(-1 / 3, 1 / 3, num=12).reshape((3, 4))

    exp_output_without_quant = jnp.matmul(inputs, float_weights)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = float_weights
    state = flax.core.freeze(state)
    outputs_with_quant = model.apply(state, inputs, padding_mask=None)
    onp.testing.assert_raises(AssertionError, onp.testing.assert_array_equal,
                              outputs_with_quant, exp_output_without_quant)
    test_utils.assert_all_close_prec(exp_output_without_quant,
                                     outputs_with_quant, weight_prec)

  # TODO(wanglisa): Add tests with bigger matrices.

  @parameterized.named_parameters(
      dict(
          testcase_name='dense_quant_fp143_scaled',
          weight_prec=fp143_scaled,
          weight_scale=onp.array([1, 2, 4, 8]),
      ),
      dict(
          testcase_name='dense_quant_fp143',
          weight_prec=fp143_unscaled,
          weight_scale=onp.array([1, 2, 4, 8]),
      ),
      dict(
          testcase_name='dense_quant_8bit',
          weight_prec=8,
          weight_scale=onp.array([1, 2, 4, 8])),
      dict(
          testcase_name='dense_quant_4bit',
          weight_prec=4,
          weight_scale=onp.array([1, 2, 4, 8])),
      dict(
          testcase_name='dense_quant_2bit',
          weight_prec=2,
          weight_scale=onp.array([1, 2, 4, 8])),
  )
  def test_weight_invariance_to_power_of_2_weight_scaling(
      self, weight_prec, weight_scale):
    # Scaling the weights before quantization by a power of 2 per channel should
    # also scale the output exactly by the same scale.

    inputs = random.uniform(self.rng_key, shape=(2, 3))
    model, state = self.init_model_with_1_layer(
        inputs, num_features=4, weight_prec=weight_prec)
    weights = random.uniform(self.rng_key, shape=(3, 4))
    weight_scale = weight_scale[jnp.newaxis, :]
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = weights
    state = flax.core.freeze(state)
    outputs_without_scaling = model.apply(state, inputs, padding_mask=None)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = jnp.multiply(weights, weight_scale)
    state = flax.core.freeze(state)
    outputs_with_scaling = model.apply(state, inputs, padding_mask=None)

    onp.testing.assert_array_equal(outputs_without_scaling * weight_scale,
                                   outputs_with_scaling)

  def test_1_bit_makes_all_weight_equal_to_zero(self):
    inputs = random.uniform(self.rng_key, shape=(2, 3))
    model, state = self.init_model_with_1_layer(
        inputs, num_features=4, weight_prec=1)
    weights = random.uniform(
        self.rng_key, shape=state['params']['kernel'].shape)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = weights
    state = flax.core.freeze(state)
    outputs = model.apply(state, inputs, padding_mask=None)
    onp.testing.assert_array_equal(outputs, onp.zeros((2, 4)))

  # TODO(shivaniagrawal): change mock tests to check for QuantOps than
  # primitives.
  @parameterized.named_parameters(
      dict(
          testcase_name='dense_quant_8bit',
          weight_prec=8,
          acts_prec=None,
          fixed_bounds=True),
      dict(
          testcase_name='dense_quant_4bit',
          weight_prec=4,
          acts_prec=None,
          fixed_bounds=True),
      dict(
          testcase_name='dense_quant_2bit',
          weight_prec=2,
          acts_prec=None,
          fixed_bounds=True),
      dict(
          testcase_name='dense_signed_input_quant_8bit',
          weight_prec=None,
          acts_prec=8,
          fixed_bounds=True),
      dict(
          testcase_name='dense_signed_input_quant_4bit',
          weight_prec=None,
          acts_prec=4,
          fixed_bounds=True),
      dict(
          testcase_name='dense_signed_input_auto_quant_8bit',
          weight_prec=None,
          acts_prec=8,
          fixed_bounds=False),
      dict(
          testcase_name='dense_signed_input_auto_quant_4bit',
          weight_prec=None,
          acts_prec=4,
          fixed_bounds=False),
      dict(
          testcase_name='dense_signed_input_quant_2bit',
          weight_prec=None,
          acts_prec=2,
          fixed_bounds=True),
  )
  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_quantized_weights_and_symmetrics_acts_should_call_clip_and_round(
      self, floor_with_gradient, round_with_gradient, weight_prec, acts_prec,
      fixed_bounds):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x

    if fixed_bounds:
      bounds = 6.0
    else:
      bounds = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        prec=acts_prec,
        bounds=bounds,
        half_shift=False)
    num_features = 4
    inputs = jnp.ones((2, 3), dtype=jnp.float32)
    model, state = self.init_model_with_1_layer(
        inputs,
        num_features,
        weight_prec=weight_prec,
        quant_act=quant_act,
        weight_half_shift=False)

    round_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(round_with_gradient.call_count, 1)
    floor_with_gradient.assert_not_called()

    round_with_gradient.reset_mock()
    floor_with_gradient.reset_mock()

    outputs = model.apply(state, inputs, padding_mask=None)

    self.assertEqual(outputs.shape, (inputs.shape[0], num_features))
    round_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(round_with_gradient.call_count, 1)
    floor_with_gradient.assert_not_called()

  # TODO(shivaniagrawal): change mock tests to check for QuantOps than
  # primitives.
  @parameterized.named_parameters(
      dict(
          testcase_name='dense_pos_quant_8bit',
          pos_inputs_prec=8,
          fixed_bounds=True),
      dict(
          testcase_name='dense_pos_quant_4bit',
          pos_inputs_prec=4,
          fixed_bounds=True),
      dict(
          testcase_name='dense_pos_quant_8bit_auto_clip',
          pos_inputs_prec=8,
          fixed_bounds=False),
      dict(
          testcase_name='dense_pos_quant_4bit_aut_clip',
          pos_inputs_prec=4,
          fixed_bounds=False),
  )
  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_quantized_inputs_should_call_clip_and_round(self,
                                                       floor_with_gradient,
                                                       round_with_gradient,
                                                       pos_inputs_prec,
                                                       fixed_bounds):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x
    if fixed_bounds:
      bounds = 6.0
    else:
      bounds = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.POSITIVE,
        prec=pos_inputs_prec,
        bounds=bounds,
        half_shift=False)
    inputs = jnp.ones((2, 3), dtype=jnp.float32)
    model, init_state = self.init_model_with_1_layer(
        inputs,
        num_features=4,
        weight_prec=None,
        quant_act=quant_act,
        weight_half_shift=False)
    floor_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(floor_with_gradient.call_count, 1)
    round_with_gradient.assert_not_called()

    round_with_gradient.reset_mock()
    floor_with_gradient.reset_mock()

    model.apply(init_state, inputs, padding_mask=None)

    floor_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(floor_with_gradient.call_count, 1)
    round_with_gradient.assert_not_called()

  @parameterized.named_parameters(
      dict(
          testcase_name='dense_quant_fp143_scaled',
          inputs_prec=fp143_scaled,
          fixed_bounds=True,
      ),)
  @mock.patch.object(fp_cast, 'downcast_sat_ftz')
  def test_fp_quantized_inputs_should_call_downcast_sat_ftz(
      self, downcast_mock, inputs_prec, fixed_bounds):

    downcast_mock.side_effect = lambda x, *_: x
    if fixed_bounds:
      bounds = 6.0
    else:
      bounds = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.POSITIVE,
        prec=inputs_prec,
        bounds=bounds,
        half_shift=False)
    inputs = jnp.ones((2, 3), dtype=jnp.float32)
    model, init_state = self.init_model_with_1_layer(
        inputs,
        num_features=4,
        weight_prec=None,
        quant_act=quant_act,
        weight_half_shift=False)
    downcast_mock.assert_called_once_with(
        mock.ANY,
        inputs_prec.fp_spec.exp_min,
        inputs_prec.fp_spec.exp_max,
        inputs_prec.fp_spec.sig_bits,
    )
    downcast_mock.reset_mock()

    model.apply(init_state, inputs, padding_mask=None)

    downcast_mock.assert_called_once_with(
        mock.ANY,
        inputs_prec.fp_spec.exp_min,
        inputs_prec.fp_spec.exp_max,
        inputs_prec.fp_spec.sig_bits,
    )

  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_without_quantized_weights_should_not_call_quantization_ops(
      self, floor_with_gradient, round_with_gradient):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x
    inputs = jnp.ones((2, 3), dtype=jnp.float32)
    model, state = self.init_model_with_1_layer(inputs, num_features=4)
    _ = model.apply(state, inputs, padding_mask=None)
    round_with_gradient.assert_not_called()
    floor_with_gradient.assert_not_called()

  @parameterized.parameters(
      dict(granularity=quant_config.QuantGranularity.PER_CHANNEL, axis=(0,)),
      dict(granularity=quant_config.QuantGranularity.PER_TENSOR, axis=None))
  @mock.patch.object(quantization, 'quantized_dot_general')
  @mock.patch.object(shape_utils, 'assert_shapes_equal')
  def test_quant_granularity(self, _, mock_quantized_dot_general, granularity,
                             axis):
    hparams = flax_layers.DenseAqt.HParams(
        weight_prec=8,
        quant_act=None,
        quant_type=quantization.QuantType.FAKE_QUANT,
        weight_quant_granularity=granularity,
        weight_half_shift=False)
    layer = flax_layers.DenseAqt(
        features=2,
        hparams=hparams,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, collect_acts_stats=False),
        paxis_name=None,
        train=False,
        dtype=jnp.float32)
    x = jnp.ones((2, 2))
    state = layer.init(self.rng_key, x, padding_mask=None)
    layer.apply(state, x, padding_mask=None)
    weight_params = mock_quantized_dot_general.call_args[1]['weight_params']
    self.assertEqual(weight_params.axis, axis)

  @parameterized.named_parameters(
      dict(
          testcase_name='train_quant',
          train=True,
          possibly_use_quantized_vars=True,
          param_info={
              'kernel': ['float32', 'embed=3', 'mlp=3'],
              'qkernel': ['float32', 'embed=3', 'mlp=3'],
              'qscale': ['float32', 'embed_qscale=1', 'mlp=3']
          }),
      dict(
          testcase_name='inference_quant',
          train=False,
          possibly_use_quantized_vars=True,
          param_info={
              'qkernel': ['int8', 'embed=3', 'mlp=3'],
              'qscale': ['float32', 'embed_qscale=1', 'mlp=3']
          }),
      dict(
          testcase_name='train_without_quant',
          train=True,
          possibly_use_quantized_vars=False,
          param_info={'kernel': ['float32', 'embed=3', 'mlp=3']}),
      dict(
          testcase_name='inference_without_quant',
          train=True,
          possibly_use_quantized_vars=False,
          param_info={'kernel': ['float32', 'embed=3', 'mlp=3']}),
  )
  def test_train_inference_differentiation(self, train,
                                           possibly_use_quantized_vars,
                                           param_info):
    num_features = 3
    inputs = random.uniform(self.rng_key, shape=(2, 3))

    _, state = self.init_model_with_1_layer(
        inputs,
        num_features=num_features,
        weight_prec=8,
        train=train,
        possibly_use_quantized_vars=possibly_use_quantized_vars,
        kernel_axis_names=('embed', 'mlp'),
        quant_type=QuantType.AQT)

    self.assertDictEqual(
        param_dtypes_shapes_axes(state['params'], state['params_axes']),
        param_info)


class EmbedLayerTest(parameterized.TestCase):
  """Tests for AQT Embed layer."""

  # TODO(shivaniagrawal): we are not raising error on jax rank
  # promotion. For EmbedAqt tests; in AQT style inputs and output are not be
  # of same shape; require more work to avoid rank promotion.
  @parameterized.named_parameters(
      dict(
          testcase_name='8_bit',
          weight_prec=8,
      ),
      dict(
          testcase_name='4_bit',
          weight_prec=4,
      ),
      dict(
          testcase_name='no_quantization',
          weight_prec=None,
      ),
  )
  def test_embed(self, weight_prec):
    # Since the dummy embedding matrix has a row of all zeros, we need 'epsilon'
    # to be added to it before calculating scale factors.
    quantization.DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING = False
    rng = random.PRNGKey(0)
    x = jnp.arange(4)[None]
    dummy_embedding = jnp.broadcast_to(jnp.arange(4)[..., None],
                                       (4, 3)).astype(jnp.float32)
    embed_module = flax_layers.EmbedAqt(
        num_embeddings=4,
        features=3,
        dtype=jnp.float32,
        hparams=flax_layers.EmbedAqt.HParams(
            weight_prec=weight_prec,
            quant_act=None,
            quant_type=QuantType.FAKE_QUANT,
            weight_half_shift=False),
        embedding_init=lambda _rng, _shape: dummy_embedding,
        train=False,
        paxis_name=None,
        dynamic_context=quant_config.DynamicContext(update_bounds=False),
    )
    y, state = embed_module.init_with_output(rng, x)
    test_utils.assert_all_close_prec(dummy_embedding[None], y, weight_prec)

    z = embed_module.apply(
        state, jnp.ones((1, 3)), padding_mask=None, method=embed_module.attend)
    test_utils.assert_all_close_prec(3. * jnp.arange(4), z[0, ...], weight_prec)

  @parameterized.named_parameters(
      dict(
          testcase_name='8_bit',
          weight_prec=8,
      ),
      dict(
          testcase_name='4_bit',
          weight_prec=4,
      ),
      dict(
          testcase_name='no_quantization',
          weight_prec=None,
      ),
  )
  def test_embed_equality(self, weight_prec):
    rng = random.PRNGKey(0)
    x = 2 * jnp.ones(4, dtype=jnp.int32)[None]
    dummy_embedding = 2 * jnp.ones((4, 2)).astype(jnp.float32)
    embed_module = flax_layers.EmbedAqt(
        num_embeddings=4,
        features=2,
        dtype=jnp.float32,
        hparams=flax_layers.EmbedAqt.HParams(
            weight_prec=weight_prec,
            quant_act=None,
            quant_type=QuantType.FAKE_QUANT,
            weight_half_shift=False),
        embedding_init=lambda _rng, _shape: dummy_embedding,
        train=False,
        dynamic_context=quant_config.DynamicContext(update_bounds=False),
        paxis_name=None)
    y, init_state = embed_module.init_with_output(rng, x)
    onp.testing.assert_array_equal(dummy_embedding[None], y)

    z = embed_module.apply(
        init_state,
        jnp.ones((1, 2)),
        padding_mask=None,
        method=embed_module.attend)
    onp.testing.assert_array_equal(2. * (2 * jnp.ones(4)), z[0, ...])

  @parameterized.named_parameters(
      dict(
          testcase_name='embed_quant_8bit',
          weight_prec=8,
          acts_prec=None,
          fixed_bounds=True),
      dict(
          testcase_name='embed_quant_4bit',
          weight_prec=4,
          acts_prec=None,
          fixed_bounds=True),
      dict(
          testcase_name='embed_input_quant_8bit',
          weight_prec=None,
          acts_prec=8,
          fixed_bounds=True),
      dict(
          testcase_name='embed_input_quant_4bit',
          weight_prec=None,
          acts_prec=4,
          fixed_bounds=True),
      dict(
          testcase_name='embed_input_auto_quant_8bit',
          weight_prec=None,
          acts_prec=8,
          fixed_bounds=False),
      dict(
          testcase_name='embed_input_auto_quant_4bit',
          weight_prec=None,
          acts_prec=4,
          fixed_bounds=False),
  )
  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_embed_should_call_clip_and_round(self, floor_with_gradient,
                                            round_with_gradient, weight_prec,
                                            acts_prec, fixed_bounds):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x

    if fixed_bounds:
      bounds = 6.0
    else:
      bounds = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        prec=acts_prec,
        bounds=bounds,
        half_shift=False)
    rng = random.PRNGKey(0)
    x = jnp.ones((1, 3))

    embed_module = flax_layers.EmbedAqt(
        num_embeddings=4,
        features=3,
        dtype=jnp.float32,
        hparams=flax_layers.EmbedAqt.HParams(
            weight_prec=weight_prec,
            quant_act=quant_act,
            quant_type=QuantType.FAKE_QUANT,
            weight_half_shift=False),
        dynamic_context=quant_config.DynamicContext(update_bounds=False),
        paxis_name=None,
        train=False)
    init_state = embed_module.init(
        rng, x, method=embed_module.attend, padding_mask=None)
    round_with_gradient.reset_mock()
    floor_with_gradient.reset_mock()
    embed_module.apply(
        init_state, x, padding_mask=None, method=embed_module.attend)
    round_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(round_with_gradient.call_count, 1)
    floor_with_gradient.assert_not_called()


class LayerNormTest(parameterized.TestCase):

  @classmethod
  def make_hparams(cls, quantize_reductions, exp_min, exp_max, sig_bits):
    prec = QuantOps.FloatQuant.FloatPrec(
        exp_min=exp_min, exp_max=exp_max, sig_bits=sig_bits)
    reduction_prec = prec if quantize_reductions else None
    hparams = flax_layers.LayerNormAqt.HParams(
        quant_hparams=flax_layers.LayerNormAqt.QuantHParams(
            prec=prec,
            reduction_prec=reduction_prec,
        ))
    return hparams

  def setUp(self):
    super().setUp()
    self.rng = jax.random.PRNGKey(0)

  @parameterized.parameters(itertools.product((False, True), (False, True)))
  def test_quantized_layer_norm_matches_unquantized_in_fp32(
      self, quantize_acts, quantize_reductions):
    # We 'quantize' to a custom floating-point format that is approximately
    # equivalent to IEEE float32 and test that results are the same as using
    # Flax's upstream unquantized LayerNorm.
    hparams = self.make_hparams(
        exp_min=-2**7,
        exp_max=2**7,
        sig_bits=23,
        quantize_reductions=quantize_reductions)
    quantized_layer_norm = flax_layers.LayerNormAqt(
        hparams=hparams,
        dtype=jnp.float32,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, quantize_acts=quantize_acts))
    x_rng, param_rng = jax.random.split(self.rng)
    x = jax.random.normal(x_rng, (3, 5))
    initial_params = quantized_layer_norm.init(param_rng, x)
    y_quantized = quantized_layer_norm.apply(initial_params, x)
    unquantized_layer_norm = nn.LayerNorm()
    y_unquantized = unquantized_layer_norm.apply(initial_params, x)
    onp.testing.assert_allclose(y_quantized, y_unquantized, rtol=2e-6)

  def test_epsilon_rounding(self):
    # We give LayerNorm a constant input. Since that input has a variance of
    # zero, we would expect layernorm to return NaN (0/0) unless the 'epsilon'
    # parameter which nudges the denominator away from zero was having an
    # effect. We test the case where the default epsilon value of 1e-6 would
    # ordinarily flush to zero after quantization with a high value of exp_min.
    # This test makes sure our code to round epsilon up to the smallest non-zero
    # representable value is wokring.
    hparams = self.make_hparams(
        exp_min=-2**2, exp_max=2**7, sig_bits=23, quantize_reductions=False)
    layer_norm = flax_layers.LayerNormAqt(
        hparams=hparams,
        use_bias=False,
        use_scale=False,
        epsilon=1e-6,
        dtype=jnp.float32,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, quantize_acts=True))
    x = jnp.ones((2, 5))
    y = layer_norm.apply({}, x)
    onp.testing.assert_equal(onp.array(y), onp.zeros(x.shape))


def assert_same_tree(a, b):
  jax.tree_map(
      functools.partial(onp.testing.assert_allclose, atol=1e-6, rtol=1e-6), a,
      b)


class DenseGeneralAqtTest(parameterized.TestCase):
  """Tests for DenseGeneralAqt layer."""

  def setUp(self):
    super(DenseGeneralAqtTest, self).setUp()
    test_utils.configure_jax()
    quantization.DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING = True
    self.rng_key = random.PRNGKey(0)

  def tearDown(self):
    config.update('jax_numpy_rank_promotion', 'warn')
    super(DenseGeneralAqtTest, self).tearDown()

  def init_model_with_1_layer(self,
                              inputs,
                              num_features,
                              axis=-1,
                              kernel_init=flax_layers.default_kernel_init,
                              weight_prec=None,
                              weight_half_shift=False,
                              quant_act=None,
                              train=False,
                              kernel_axis_names=None,
                              reshape_kernel=True,
                              possibly_use_quantized_vars=False):
    """Create and initialize a flax model with a single DenseAqt layer."""
    layer_kwargs = {
        'kernel_init': kernel_init,
        'features': num_features,
        'axis': axis,
        'use_bias': False,
        'train': train,
        'dtype': jnp.float32,
        'kernel_axis_names': kernel_axis_names,
        'reshape_kernel': reshape_kernel,
        'possibly_use_quantized_vars': possibly_use_quantized_vars,
    }
    layer_kwargs['hparams'] = flax_layers.DenseGeneralAqt.HParams(
        weight_prec=weight_prec,
        quant_act=quant_act,
        weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
        weight_half_shift=weight_half_shift)

    dense_module = flax_layers.DenseGeneralAqt(**layer_kwargs)
    initial_state = dense_module.init(self.rng_key, jnp.zeros(inputs.shape))
    return dense_module, initial_state

  # DenseGeneral does not support fq version, hence fp quantization isn't
  # supported either
  @parameterized.named_parameters(
      dict(testcase_name='float', weight_prec=None),

      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
      dict(testcase_name='quant_4bit_no_reshape', weight_prec=4, reshape=False),
      dict(testcase_name='quant_2bit', weight_prec=2),
  )
  def test_ones_weights_should_give_precise_output(self,
                                                   weight_prec,
                                                   reshape=True):
    """If all weights are 1, no quantization error should be introduced."""
    inputs = random.uniform(self.rng_key, shape=(2, 3, 4))
    axis = (1, 2)
    model, state = self.init_model_with_1_layer(
        inputs,
        num_features=5,
        axis=axis,
        kernel_init=initializers.ones,
        weight_prec=weight_prec,
        weight_half_shift=False,
        reshape_kernel=reshape)
    outputs = model.apply(state, inputs)
    contract_ind = tuple(range(0, len(axis)))
    exp_outputs = lax.dot_general(
        inputs,
        jnp.reshape(state['params']['kernel'], (3, 4, 5)),
        dimension_numbers=((axis, contract_ind), ((), ())))
    onp.testing.assert_array_almost_equal(outputs, exp_outputs)

  @parameterized.named_parameters(
      dict(testcase_name='reshape_kernel', reshape_kernel=True),
      dict(testcase_name='no_reshape', reshape_kernel=False),
  )
  def test_logical_axis_names(self, reshape_kernel):
    inputs = random.uniform(self.rng_key, shape=(2, 3, 1, 4))
    _, state = self.init_model_with_1_layer(
        inputs,
        num_features=4,
        axis=(2, 3),
        kernel_init=initializers.ones,
        weight_prec=8,
        weight_half_shift=False,
        kernel_axis_names=('embed', 'head', 'mlp'),
        reshape_kernel=reshape_kernel)

    expected_shape_axes = [
        'float32', 'embed * head=4', 'mlp=4'
    ] if reshape_kernel else ['float32', 'embed=1', 'head=4', 'mlp=4']

    self.assertDictEqual(
        param_dtypes_shapes_axes(state['params'], state['params_axes']),
        {'kernel': expected_shape_axes})

  @parameterized.named_parameters(
      dict(testcase_name='dense_quant_8bit', weight_prec=8),
      dict(testcase_name='dense_quant_4bit', weight_prec=4),
      dict(testcase_name='dense_quant_2bit', weight_prec=2),
  )
  def test_full_range_integer_weights_should_give_precise_output(
      self, weight_prec):
    # If weights are ints (already quantized) and
    # max(abs(weights[:, ch])) == 2**(prec-1)-1 in each channel,
    # no quantization error should be introduced.
    num_features = 256
    input_dim = 1024
    inputs = random.uniform(self.rng_key, shape=(2, input_dim))
    model, state = self.init_model_with_1_layer(
        inputs,
        num_features,
        axis=(1,),
        weight_prec=weight_prec,
        weight_half_shift=False)
    minval = -2**(weight_prec - 1) + 1
    maxval = 2**(weight_prec - 1) - 1

    full_range_integer_weights = random.randint(self.rng_key,
                                                (input_dim, num_features),
                                                minval, maxval + 1)

    # manually set one value in each output dim of weights to be exactly maxval
    full_range_integer_weights = full_range_integer_weights.at[0, :].set(maxval)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = full_range_integer_weights
    state = flax.core.freeze(state)
    outputs = model.apply(state, inputs)
    exp_outputs = jnp.matmul(inputs, state['params']['kernel'])
    onp.testing.assert_array_equal(outputs, exp_outputs)

  @parameterized.named_parameters(
      dict(
          testcase_name='densegeneral_quant_8bit',
          weight_prec=8,
          expected_output=[[-0.08770747, -0.00193817, 0.08321345, 0.16902271],
                           [-0.2838632, -0.1654416, -0.04855134, 0.06976502]]),
      dict(
          testcase_name='densegeneral_quant_4bit',
          weight_prec=4,
          expected_output=[[-0.09499891, -0.016027, 0.09432255, 0.17474438],
                           [-0.29012504, -0.17881386, -0.02733358,
                            0.08016007]]),
      dict(
          testcase_name='densegeneral_quant_2bit',
          weight_prec=2,
          expected_output=[[-0.19150016, 0.0456724, 0.0456724, 0.24732196],
                           [-0.29805943, -0.12025267, -0.12025267,
                            0.15108395]]),
  )
  def test_float_weights_regression_test(self, weight_prec, expected_output):
    inputs = random.uniform(self.rng_key, shape=(2, 3))
    model, state = self.init_model_with_1_layer(
        inputs, num_features=4, weight_prec=weight_prec)
    float_weights = jnp.linspace(-1 / 3, 1 / 3, num=12).reshape((3, 4))

    exp_output_without_quant = jnp.matmul(inputs, float_weights)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = float_weights
    state = flax.core.freeze(state)
    outputs_with_quant = model.apply(state, inputs)
    onp.testing.assert_raises(AssertionError, onp.testing.assert_array_equal,
                              outputs_with_quant, exp_output_without_quant)
    onp.testing.assert_allclose(outputs_with_quant, expected_output, rtol=1e-5)
    onp.testing.assert_allclose(
        exp_output_without_quant,
        [[-0.08814345, -0.00231601, 0.0835114, 0.16933881],
         [-0.2836182, -0.16580023, -0.04798227, 0.06983571]],
        rtol=1e-5)
    test_utils.assert_all_close_prec(exp_output_without_quant,
                                     outputs_with_quant, weight_prec)

  @parameterized.named_parameters(
      # TODO(shivaniagrawal): this test is flaky and fails with rtol=0.0004
      # with given rtol=0.0001
      # dict(
      #     testcase_name='dense_quant_8bit',
      #     layer_class=flax_layers.DenseAqt,
      #     weight_prec=8),
      dict(testcase_name='dense_quant_4bit', weight_prec=4),
      dict(testcase_name='dense_quant_2bit', weight_prec=2),
  )
  def test_full_range_integer_weights_with_float_scale_should_give_close_output(
      self, weight_prec):
    # If weights are ints (already quantized) with
    # max(abs(weights[:, ch])) == 2**(prec-1)-1 in each channel
    # and if these integer weights are multiplied by a float scale,
    # the resulting error should still be very small (just float rounding).

    num_features = 256
    input_dim = 1024
    inputs = random.uniform(self.rng_key, shape=(2, input_dim))
    model, state = self.init_model_with_1_layer(
        inputs, num_features, axis=(1,), weight_prec=weight_prec)
    minval = -2**(weight_prec - 1) + 1
    maxval = 2**(weight_prec - 1) - 1

    full_range_integer_weights = random.randint(self.rng_key,
                                                (input_dim, num_features),
                                                minval, maxval + 1)

    # manually set one value in each output dim of weights to be exactly maxval
    full_range_integer_weights = full_range_integer_weights.at[0, :].set(maxval)

    float_scale = jax.random.uniform(self.rng_key, (1, num_features))
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = full_range_integer_weights * float_scale
    state = flax.core.freeze(state)
    outputs = model.apply(state, inputs)
    exp_outputs = jnp.matmul(inputs, state['params']['kernel'])
    # TODO(wanglisa): Determine how much noise is expected for following test.
    # We know that the noise should be proportional to the square root of
    # input_dim and inversely proportional to 2**weight_prec.
    # The following tol_const was obtained experimentally and should be derived
    # more systematically.
    tol_const = 8e-04
    onp.testing.assert_allclose(
        outputs,
        exp_outputs,
        rtol=jnp.sqrt(input_dim) * 2**(-weight_prec) * tol_const)

  @parameterized.named_parameters(
      dict(testcase_name='dense_quant_4bit', weight_prec=4),
      dict(testcase_name='dense_quant_2bit', weight_prec=2),
  )
  def test_float_weights_should_give_close_output(self, weight_prec):
    inputs = random.uniform(self.rng_key, shape=(2, 3))
    model, state = self.init_model_with_1_layer(
        inputs, num_features=4, weight_prec=weight_prec)
    float_weights = jnp.linspace(-1 / 3, 1 / 3, num=12).reshape((3, 4))

    exp_output_without_quant = jnp.matmul(inputs, float_weights)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = float_weights
    state = flax.core.freeze(state)
    outputs_with_quant = model.apply(state, inputs)
    onp.testing.assert_raises(AssertionError, onp.testing.assert_array_equal,
                              outputs_with_quant, exp_output_without_quant)

    test_utils.assert_all_close_prec(exp_output_without_quant,
                                     outputs_with_quant, weight_prec)

  @parameterized.named_parameters(
      dict(
          testcase_name='dense_quant_8bit',
          weight_prec=8,
          weight_scale=onp.array([1, 2, 4, 8])),
      dict(
          testcase_name='dense_quant_4bit',
          weight_prec=4,
          weight_scale=onp.array([1, 2, 4, 8])),
      dict(
          testcase_name='dense_quant_2bit',
          weight_prec=2,
          weight_scale=onp.array([1, 2, 4, 8])),
  )
  def test_weight_invariance_to_power_of_2_weight_scaling(
      self, weight_prec, weight_scale):
    # Scaling the weights before quantization by a power of 2 per channel should
    # also scale the output exactly by the same scale.

    inputs = random.uniform(self.rng_key, shape=(2, 3))
    model, state = self.init_model_with_1_layer(
        inputs, num_features=4, axis=(1,), weight_prec=weight_prec)
    weights = random.uniform(self.rng_key, shape=(3, 4))
    weight_scale = weight_scale[jnp.newaxis, :]
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = weights
    state = flax.core.freeze(state)
    outputs_without_scaling = model.apply(state, inputs)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = jnp.multiply(weights, weight_scale)
    state = flax.core.freeze(state)
    outputs_with_scaling = model.apply(state, inputs)

    onp.testing.assert_array_equal(outputs_without_scaling * weight_scale,
                                   outputs_with_scaling)

  def test_1_bit_makes_all_weight_equal_to_zero(self):
    inputs = random.uniform(self.rng_key, shape=(2, 3))
    model, state = self.init_model_with_1_layer(
        inputs, num_features=4, axis=(1,), weight_prec=1)
    weights = random.uniform(
        self.rng_key, shape=state['params']['kernel'].shape)
    state = flax.core.unfreeze(state)
    state['params']['kernel'] = weights
    state = flax.core.freeze(state)
    outputs = model.apply(state, inputs)
    onp.testing.assert_array_equal(outputs, onp.zeros((2, 4)))

  # TODO(shivaniagrawal): change mock tests to check for QuantOps than
  # primitives.
  @parameterized.named_parameters(
      dict(
          testcase_name='dense_weight_quant_8bit',
          weight_prec=8,
          acts_prec=None),
      dict(
          testcase_name='dense_weight_quant_4bit',
          weight_prec=4,
          acts_prec=None),
      dict(
          testcase_name='dense_weight_quant_2bit',
          weight_prec=2,
          acts_prec=None),
      dict(
          testcase_name='dense_signed_input_quant_8bit',
          weight_prec=None,
          acts_prec=8),
      dict(
          testcase_name='dense_signed_input_quant_4bit',
          weight_prec=None,
          acts_prec=4),
      dict(
          testcase_name='dense_signed_input_quant_2bit',
          weight_prec=None,
          acts_prec=2),
      dict(
          testcase_name='dense_signed_input_auto_quant_8bit',
          weight_prec=None,
          acts_prec=8,
          fixed_bounds=False),
      dict(
          testcase_name='dense_signed_input_auto_quant_4bit',
          weight_prec=None,
          acts_prec=4,
          fixed_bounds=False),
  )
  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_quantized_weights_and_symmetrics_acts_should_call_clip_and_round(
      self,
      floor_with_gradient,
      round_with_gradient,
      weight_prec,
      acts_prec,
      fixed_bounds=False):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x

    if fixed_bounds:
      bounds = 6.0
    else:
      bounds = get_bounds.DynamicBounds.Hyper(
          granularity=quant_config.QuantGranularity.PER_CHANNEL)
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        prec=acts_prec,
        bounds=bounds,
        half_shift=False)
    num_features = 4
    inputs = jnp.ones((2, 3), dtype=jnp.float32)
    model, state = self.init_model_with_1_layer(
        inputs,
        num_features,
        weight_prec=weight_prec,
        weight_half_shift=False,
        quant_act=quant_act,
    )

    round_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(round_with_gradient.call_count, 1)
    floor_with_gradient.assert_not_called()

    round_with_gradient.reset_mock()
    floor_with_gradient.reset_mock()

    outputs = model.apply(state, inputs)

    self.assertEqual(outputs.shape, (inputs.shape[0], num_features))
    round_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(round_with_gradient.call_count, 1)
    floor_with_gradient.assert_not_called()

  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_without_quantized_weights_should_not_call_quantization_ops(
      self, floor_with_gradient, round_with_gradient):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x
    inputs = jnp.ones((2, 3), dtype=jnp.float32)
    model, state = self.init_model_with_1_layer(inputs, num_features=4)
    _ = model.apply(state, inputs)
    round_with_gradient.assert_not_called()
    floor_with_gradient.assert_not_called()

  # TODO(shivaniagrawal): change mock tests to check for QuantOps than
  # primitives.
  @parameterized.named_parameters(
      dict(
          testcase_name='dense_pos_quant_8bit',
          pos_inputs_prec=8,
          fixed_bounds=True),
      dict(
          testcase_name='dense_pos_quant_4bit',
          pos_inputs_prec=4,
          fixed_bounds=True),
      dict(
          testcase_name='dense_pos_quant_8bit_auto_clip',
          pos_inputs_prec=8,
          fixed_bounds=False),
      dict(
          testcase_name='dense_pos_quant_4bit_auto_clip',
          pos_inputs_prec=4,
          fixed_bounds=False),
  )
  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_quantized_inputs_should_call_clip_and_round(self,
                                                       floor_with_gradient,
                                                       round_with_gradient,
                                                       pos_inputs_prec,
                                                       fixed_bounds):

    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x
    if fixed_bounds:
      bounds = 6.0
    else:
      bounds = get_bounds.DynamicBounds.Hyper(
          granularity=quant_config.QuantGranularity.PER_CHANNEL)
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.POSITIVE,
        prec=pos_inputs_prec,
        bounds=bounds,
        half_shift=False)
    inputs = jnp.ones((2, 3), dtype=jnp.float32)
    model, init_state = self.init_model_with_1_layer(
        inputs,
        num_features=4,
        weight_prec=None,
        quant_act=quant_act,
        weight_half_shift=False)
    floor_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(floor_with_gradient.call_count, 1)
    round_with_gradient.assert_not_called()

    round_with_gradient.reset_mock()
    floor_with_gradient.reset_mock()

    model.apply(init_state, inputs)

    floor_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(floor_with_gradient.call_count, 1)
    round_with_gradient.assert_not_called()

  @parameterized.parameters(
      dict(granularity=quant_config.QuantGranularity.PER_CHANNEL, axis=(0, 1)),
      dict(granularity=quant_config.QuantGranularity.PER_TENSOR, axis=None))
  @mock.patch.object(quantization, 'flaxformer_dot_general')
  @mock.patch.object(shape_utils, 'assert_shapes_equal')
  def test_weights_quant_granularity(self, _, mock_flaxformer_dot_general,
                                     granularity, axis):
    hparams = flax_layers.DenseGeneralAqt.HParams(
        weight_prec=8,
        quant_act=None,
        weight_quant_granularity=granularity,
        weight_half_shift=False)
    layer = flax_layers.DenseGeneralAqt(
        features=3,
        axis=(1, 2),
        hparams=hparams,
        train=False,
        use_bias=False,
        dtype=jnp.float32)
    x = jnp.ones((2, 3, 4))
    state = layer.init(self.rng_key, x)
    layer.apply(state, x)
    weight_params = mock_flaxformer_dot_general.call_args[1]['weight_params']
    self.assertEqual(weight_params.axis, axis)

  @parameterized.parameters(
      dict(
          granularity=quant_config.QuantGranularity.PER_CHANNEL,
          quant_axis=(1, 2)),
      dict(
          granularity=quant_config.QuantGranularity.PER_TENSOR,
          quant_axis=None))
  @mock.patch.object(quantization, 'flaxformer_dot_general')
  @mock.patch.object(shape_utils, 'assert_shapes_equal')
  def test_acts_quant_granularity(self, _, mock_flaxformer_dot_general,
                                  granularity, quant_axis):
    bounds = get_bounds.DynamicBounds.Hyper(granularity=granularity)
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.POSITIVE,
        prec=8,
        bounds=bounds,
        half_shift=False)
    hparams = flax_layers.DenseGeneralAqt.HParams(
        weight_prec=None,
        quant_act=quant_act,
        weight_quant_granularity=quant_config.QuantGranularity.PER_TENSOR,
        weight_half_shift=False)
    layer = flax_layers.DenseGeneralAqt(
        features=2,
        axis=(1, 2),
        hparams=hparams,
        train=False,
        use_bias=False,
        dtype=jnp.float32)
    x = jnp.ones((2, 3, 4))
    state = layer.init(self.rng_key, x)
    layer.apply(state, x)
    bounds_params = mock_flaxformer_dot_general.call_args[1]['bounds_params']
    self.assertEqual(bounds_params.quant_axis, quant_axis)

  @parameterized.named_parameters(
      dict(
          testcase_name='train_quant',
          train=True,
          possibly_use_quantized_vars=True,
          param_info={
              'kernel': ['float32', 'embed=3', 'mlp=3'],
              'qkernel': ['float32', 'embed=3', 'mlp=3'],
              'qscale': ['float32', 'embed_qscale=1', 'mlp=3']
          }),
      dict(
          testcase_name='inference_quant',
          train=False,
          possibly_use_quantized_vars=True,
          param_info={
              'qkernel': ['int8', 'embed=3', 'mlp=3'],
              'qscale': ['float32', 'embed_qscale=1', 'mlp=3']
          }),
      dict(
          testcase_name='train_without_quant',
          train=True,
          possibly_use_quantized_vars=False,
          param_info={'kernel': ['float32', 'embed=3', 'mlp=3']}),
      dict(
          testcase_name='inference_without_quant',
          train=True,
          possibly_use_quantized_vars=False,
          param_info={'kernel': ['float32', 'embed=3', 'mlp=3']}),
  )
  def test_train_inference_differentiation(self, train,
                                           possibly_use_quantized_vars,
                                           param_info):
    num_features = 3
    inputs = random.uniform(self.rng_key, shape=(2, 3))

    _, state = self.init_model_with_1_layer(
        inputs,
        num_features=num_features,
        weight_prec=8,
        train=train,
        possibly_use_quantized_vars=possibly_use_quantized_vars,
        kernel_axis_names=('embed', 'mlp'))

    self.assertDictEqual(
        param_dtypes_shapes_axes(state['params'], state['params_axes']),
        param_info)


class DenseGeneralTest(parameterized.TestCase):

  # pylint: disable=unused-argument
  def _mock_initializer(self, key, shape, dtype=jnp.float_, val=1.0):  # pylint: disable=g-unreachable-test-method
    return jnp.ones(shape, dtypes.canonicalize_dtype(dtype)) * val
  # pylint: enable=unused-argument

  def setUp(self):
    super(DenseGeneralTest, self).setUp()
    self.hparams = flax_layers.DenseGeneralAqt.HParams(
        weight_prec=None,
        quant_act=None,
        weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
        weight_half_shift=False)

  def test_dense_general_no_bias(self):
    rng = random.PRNGKey(0)
    x = jnp.ones((1, 3))
    model = flax_layers.DenseGeneralAqt(
        hparams=self.hparams,
        train=False,
        features=4,
        use_bias=False,
        kernel_init=initializers.ones,
    )
    y, _ = model.init_with_output(rng, x)
    self.assertEqual(y.shape, (1, 4))
    onp.testing.assert_allclose(y, onp.full((1, 4), 3.))

  def test_dense_general_with_bias(self):
    rng = random.PRNGKey(0)
    x = jnp.ones((1, 3))
    model = flax_layers.DenseGeneralAqt(
        hparams=self.hparams,
        train=False,
        features=4,
        use_bias=True,
        kernel_init=initializers.ones,
        bias_init=initializers.ones,
    )
    y, _ = model.init_with_output(rng, x)
    self.assertEqual(y.shape, (1, 4))
    onp.testing.assert_allclose(y, onp.full((1, 4), 4.))

  def test_dense_general_two_features(self):
    rng = random.PRNGKey(0)
    x = jnp.ones((1, 3))
    model = flax_layers.DenseGeneralAqt(
        hparams=self.hparams,
        train=False,
        features=(2, 2),
        use_bias=False,
        kernel_init=initializers.ones,
        bias_init=initializers.ones,
        kernel_axis_names=('a', 'b', 'c'),
    )
    y, variables = model.init_with_output(rng, x)
    # We transform the last input dimension to two output dimensions (2, 2).
    onp.testing.assert_allclose(y, onp.full((1, 2, 2), 3.))


  def test_dense_general_two_axes(self):
    rng = random.PRNGKey(0)
    x = jnp.ones((1, 2, 2))
    model = flax_layers.DenseGeneralAqt(
        hparams=self.hparams,
        train=False,
        features=3,
        use_bias=False,
        axis=(-2, 2),  # Note: this is the same as (1, 2).
        kernel_init=initializers.ones,
        bias_init=initializers.ones,
        kernel_axis_names=('a', 'b', 'c'),
    )
    y, variables = model.init_with_output(rng, x)
    # We transform the last two input dimensions (2, 2) to one output dimension.
    onp.testing.assert_allclose(y, onp.full((1, 3), 4.))



if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utilities for Jax AQTp."""

from aqt.common import aqt_config
from aqt.jax import aqt_tensor
import jax.numpy as jnp

# pylint: disable=protected-access
# pytype: disable=attribute-error


def possibly_use_quantized_variable(
    quantizer: aqt_tensor.TensorQuantizer,
    x: jnp.ndarray,
    train: bool) -> jnp.ndarray:
  """Returns quantized variable if not training and TQ.use_quantized_variable, casted to x.dtype.

  Overrides x with quantizer's quantized variable, cast to x.dtype, if
  quantizer.use_quantized variable, and return x otherwise. Note this will
  return the quantized variable even if a FloatConfig is used.

  For motivation behind these semantics, see b/219040448.

  Args:
    quantizer: TensorQuantizer for the input tensor x.
    x: lhs or rhs of conv_general.
    train: Indicates if in training or not.

  Returns:
    The input tensor x or its quantized one.
  """
  if quantizer.config is not None and quantizer.config.use_quantized_variable and not train:
    qx = quantizer.quantized_variable.value
    qx = qx.astype(x.dtype)
    return qx
  return x


def should_int8_quantize(
    lhs_quantizer: aqt_tensor.TensorQuantizer,
    rhs_quantizer: aqt_tensor.TensorQuantizer) -> bool:
  """Determines whether or not to quantize."""

  if lhs_quantizer.config is None or rhs_quantizer.config is None:
    return jnp.bool_(False)

  lhs_configs = lhs_quantizer.config.tensor_configs
  rhs_configs = rhs_quantizer.config.tensor_configs

  should_quantize = False
  for lhs_config in lhs_configs:
    for rhs_config in rhs_configs:
      if (isinstance(lhs_config.quant_config, aqt_config.IntQuantConfig) and
          isinstance(rhs_config.quant_config, aqt_config.IntQuantConfig) and
          lhs_config.quant_config.bits <= 8 and
          rhs_config.quant_config.bits <= 8):
        should_quantize |= (
            aqt_tensor.is_config_active(lhs_config,
                                        lhs_quantizer._last_update.value)
            & aqt_tensor.is_config_active(rhs_config,
                                          rhs_quantizer._last_update.value))

  return should_quantize



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ImageNet input pipeline."""

import jax

import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds

TRAIN_IMAGES = 1281167
EVAL_IMAGES = 50000

IMAGE_SIZE = 224
CROP_PADDING = 32
MEAN_RGB = [0.485 * 255, 0.456 * 255, 0.406 * 255]
STDDEV_RGB = [0.229 * 255, 0.224 * 255, 0.225 * 255]


def distorted_bounding_box_crop(image_bytes,
                                bbox,
                                min_object_covered=0.1,
                                aspect_ratio_range=(0.75, 1.33),
                                area_range=(0.05, 1.0),
                                max_attempts=100):
  """Generates cropped_image using one of the bboxes randomly distorted.

  See `tf.image.sample_distorted_bounding_box` for more documentation.

  Args:
    image_bytes: `Tensor` of binary image data.
    bbox: `Tensor` of bounding boxes arranged `[1, num_boxes, coords]` where
      each coordinate is [0, 1) and the coordinates are arranged as `[ymin,
      xmin, ymax, xmax]`. If num_boxes is 0 then use the whole image.
    min_object_covered: An optional `float`. Defaults to `0.1`. The cropped area
      of the image must contain at least this fraction of any bounding box
      supplied.
    aspect_ratio_range: An optional list of `float`s. The cropped area of the
      image must have an aspect ratio = width / height within this range.
    area_range: An optional list of `float`s. The cropped area of the image must
      contain a fraction of the supplied image within in this range.
    max_attempts: An optional `int`. Number of attempts at generating a cropped
      region of the image of the specified constraints. After `max_attempts`
      failures, return the entire image.

  Returns:
    cropped image `Tensor`
  """
  shape = tf.image.extract_jpeg_shape(image_bytes)
  sample_distorted_bounding_box = tf.image.sample_distorted_bounding_box(
      shape,
      bounding_boxes=bbox,
      min_object_covered=min_object_covered,
      aspect_ratio_range=aspect_ratio_range,
      area_range=area_range,
      max_attempts=max_attempts,
      use_image_if_no_bounding_boxes=True)
  bbox_begin, bbox_size, _ = sample_distorted_bounding_box

  # Crop the image to the specified bounding box.
  offset_y, offset_x, _ = tf.unstack(bbox_begin)
  target_height, target_width, _ = tf.unstack(bbox_size)
  crop_window = tf.stack([offset_y, offset_x, target_height, target_width])
  image = tf.image.decode_and_crop_jpeg(image_bytes, crop_window, channels=3)

  return image


def _resize(image, image_size):
  return tf.image.resize([image], [image_size, image_size],
                         method=tf.image.ResizeMethod.BICUBIC)[0]


def _at_least_x_are_equal(a, b, x):
  """At least `x` of `a` and `b` `Tensors` are equal."""
  match = tf.equal(a, b)
  match = tf.cast(match, tf.int32)
  return tf.greater_equal(tf.reduce_sum(match), x)


def _decode_and_random_crop(image_bytes, image_size):
  """Make a random crop of image_size."""
  bbox = tf.constant([0.0, 0.0, 1.0, 1.0], dtype=tf.float32, shape=[1, 1, 4])
  image = distorted_bounding_box_crop(
      image_bytes,
      bbox,
      min_object_covered=0.1,
      aspect_ratio_range=(3. / 4, 4. / 3.),
      area_range=(0.08, 1.0),
      max_attempts=10)
  original_shape = tf.image.extract_jpeg_shape(image_bytes)
  bad = _at_least_x_are_equal(original_shape, tf.shape(image), 3)

  image = tf.cond(bad, lambda: _decode_and_center_crop(image_bytes, image_size),
                  lambda: _resize(image, image_size))

  return image


def _decode_and_center_crop(image_bytes, image_size):
  """Crops to center of image with padding then scales image_size."""
  shape = tf.image.extract_jpeg_shape(image_bytes)
  image_height = shape[0]
  image_width = shape[1]

  padded_center_crop_size = tf.cast(
      ((image_size / (image_size + CROP_PADDING)) *
       tf.cast(tf.minimum(image_height, image_width), tf.float32)), tf.int32)

  offset_height = ((image_height - padded_center_crop_size) + 1) // 2
  offset_width = ((image_width - padded_center_crop_size) + 1) // 2
  crop_window = tf.stack([
      offset_height, offset_width, padded_center_crop_size,
      padded_center_crop_size
  ])
  image = tf.image.decode_and_crop_jpeg(image_bytes, crop_window, channels=3)
  image = _resize(image, image_size)

  return image


def normalize_image(image):
  image -= tf.constant(MEAN_RGB, shape=[1, 1, 3], dtype=image.dtype)
  image /= tf.constant(STDDEV_RGB, shape=[1, 1, 3], dtype=image.dtype)
  return image


def preprocess_for_train(image_bytes, dtype=tf.float32, image_size=IMAGE_SIZE):
  """Preprocesses the given image for training.

  Args:
    image_bytes: `Tensor` representing an image binary of arbitrary size.
    dtype: data type of the image.
    image_size: image size.

  Returns:
    A preprocessed image `Tensor`.
  """
  image = _decode_and_random_crop(image_bytes, image_size)
  image = tf.reshape(image, [image_size, image_size, 3])
  image = tf.image.random_flip_left_right(image)
  image = normalize_image(image)
  image = tf.image.convert_image_dtype(image, dtype=dtype)
  return image


def preprocess_for_eval(image_bytes, dtype=tf.float32, image_size=IMAGE_SIZE):
  """Preprocesses the given image for evaluation.

  Args:
    image_bytes: `Tensor` representing an image binary of arbitrary size.
    dtype: data type of the image.
    image_size: image size.

  Returns:
    A preprocessed image `Tensor`.
  """
  image = _decode_and_center_crop(image_bytes, image_size)
  image = tf.reshape(image, [image_size, image_size, 3])
  image = normalize_image(image)
  image = tf.image.convert_image_dtype(image, dtype=dtype)
  return image


def load_split(batch_size,
               train,
               data_dir,
               dtype=tf.float32,
               image_size=IMAGE_SIZE,
               cache=False):
  """Creates a split from the ImageNet dataset using TensorFlow Datasets.

  Args:
    batch_size: the batch size returned by the data pipeline.
    train: Whether to load the train or evaluation split.
    data_dir: str, directory to read/write data. Defaults to the value of the
      environment variable TFDS_DATA_DIR, if set, otherwise falls back to
      '~/tensorflow_datasets'.
    dtype: data type of the image.
    image_size: The target size of the images.
    cache: Whether to cache the dataset.

  Returns:
    A `tf.data.Dataset`.
  """
  if train:
    split_size = TRAIN_IMAGES // jax.host_count()
    start = jax.host_id() * split_size
    split = 'train[{}:{}]'.format(start, start + split_size)
  else:
    split_size = EVAL_IMAGES // jax.host_count()
    start = jax.host_id() * split_size
    split = 'validation[{}:{}]'.format(start, start + split_size)

  def decode_example(example):
    if train:
      image = preprocess_for_train(example['image'], dtype, image_size)
    else:
      image = preprocess_for_eval(example['image'], dtype, image_size)
    return {'image': image, 'label': example['label']}

  ds = tfds.load(
      'imagenet2012:5.*.*',
      split=split,
      data_dir=data_dir,
      decoders={
          'image': tfds.decode.SkipDecoding(),
      })
  options = tf.data.Options()
  options.experimental_threading.private_threadpool_size = 48
  ds = ds.with_options(options)

  if cache:
    ds = ds.cache()

  if train:
    ds = ds.repeat()
    ds = ds.shuffle(16 * batch_size, seed=0)

  ds = ds.map(decode_example, num_parallel_calls=tf.data.experimental.AUTOTUNE)
  ds = ds.batch(batch_size, drop_remainder=True)

  if not train:
    ds = ds.repeat()

  ds = ds.prefetch(10)

  return ds



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT configs."""
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2 import utils
import jax.numpy as jnp


class AqtConfigTest(parameterized.TestCase):

  def test_config_v4(self):
    cfg = config.config_v4(
        fwd_bits=8,
        dlhs_bits=7,
        drhs_bits=6,
        rng_type='custom-1',
        dlhs_local_aqt=config.LocalAqt(2),
        drhs_local_aqt=config.LocalAqt(3),
        fwd_accumulator_dtype=jnp.int16,
        dlhs_accumulator_dtype=jnp.int8,
        drhs_accumulator_dtype=jnp.int4,
    )
    expected_cfg_str = """DotGeneral(fwd=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=8,
                                                                                 preserve_zero=True,
                                                                                 preserve_max_val=False,
                                                                                 clip=True,
                                                                                 clip_gradient=False,
                                                                                 round=True,
                                                                                 noise_fn=None,
                                                                                 dtype=<class 'jax.numpy.int8'>),
                                                            calib_shared_axes=None,
                                                            scale_stop_grad=True,
                                                            calibration=AbsMaxCalibration(scale=None),
                                                            po2_scale=False,
                                                            context=Context(key=None,
                                                                            train_step=None)),
                                        use_fwd_quant=None,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             rhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=8,
                                                                                 preserve_zero=True,
                                                                                 preserve_max_val=False,
                                                                                 clip=True,
                                                                                 clip_gradient=False,
                                                                                 round=True,
                                                                                 noise_fn=None,
                                                                                 dtype=<class 'jax.numpy.int8'>),
                                                            calib_shared_axes=None,
                                                            scale_stop_grad=True,
                                                            calibration=AbsMaxCalibration(scale=None),
                                                            po2_scale=False,
                                                            context=Context(key=None,
                                                                            train_step=None)),
                                        use_fwd_quant=None,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             dg_accumulator_dtype=<class 'jax.numpy.int16'>,
                             local_aqt=None,
                             jax_scope_name='aqt_fwd',
                             allow_dummy_gradient_into_qtensor=False),
           dlhs=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=7,
                                                                                  preserve_zero=True,
                                                                                  preserve_max_val=False,
                                                                                  clip=True,
                                                                                  clip_gradient=False,
                                                                                  round=True,
                                                                                  noise_fn=RandomCenteredUniform(),
                                                                                  dtype=<class 'jax.numpy.int8'>),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=None,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=7,
                                                                                  preserve_zero=True,
                                                                                  preserve_max_val=False,
                                                                                  clip=True,
                                                                                  clip_gradient=False,
                                                                                  round=True,
                                                                                  noise_fn=None,
                                                                                  dtype=<class 'jax.numpy.int8'>),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_accumulator_dtype=<class 'jax.numpy.int8'>,
                              local_aqt=LocalAqt(contraction_axis_shard_count=2),
                              jax_scope_name='aqt_dlhs',
                              allow_dummy_gradient_into_qtensor=False),
           drhs=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=6,
                                                                                  preserve_zero=True,
                                                                                  preserve_max_val=False,
                                                                                  clip=True,
                                                                                  clip_gradient=False,
                                                                                  round=True,
                                                                                  noise_fn=RandomCenteredUniform(),
                                                                                  dtype=<class 'jax.numpy.int8'>),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=None,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=6,
                                                                                  preserve_zero=True,
                                                                                  preserve_max_val=False,
                                                                                  clip=True,
                                                                                  clip_gradient=False,
                                                                                  round=True,
                                                                                  noise_fn=None,
                                                                                  dtype=<class 'jax.numpy.int8'>),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_accumulator_dtype=<class 'jax.numpy.int4'>,
                              local_aqt=LocalAqt(contraction_axis_shard_count=3),
                              jax_scope_name='aqt_drhs',
                              allow_dummy_gradient_into_qtensor=False))"""
    utils.test_pprint_eq(cfg, expected_cfg_str)

  def test_configv4_original(self):
    expected_cfg_str = """DotGeneral(fwd=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=8,
                                                                                 preserve_zero=True,
                                                                                 preserve_max_val=False,
                                                                                 clip=True,
                                                                                 clip_gradient=False,
                                                                                 round=True,
                                                                                 noise_fn=None,
                                                                                 dtype=<class 'jax.numpy.int8'>),
                                                            calib_shared_axes=None,
                                                            scale_stop_grad=True,
                                                            calibration=AbsMaxCalibration(scale=None),
                                                            po2_scale=False,
                                                            context=Context(key=None,
                                                                            train_step=None)),
                                        use_fwd_quant=None,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             rhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=8,
                                                                                 preserve_zero=True,
                                                                                 preserve_max_val=False,
                                                                                 clip=True,
                                                                                 clip_gradient=False,
                                                                                 round=True,
                                                                                 noise_fn=None,
                                                                                 dtype=<class 'jax.numpy.int8'>),
                                                            calib_shared_axes=None,
                                                            scale_stop_grad=True,
                                                            calibration=AbsMaxCalibration(scale=None),
                                                            po2_scale=False,
                                                            context=Context(key=None,
                                                                            train_step=None)),
                                        use_fwd_quant=None,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             dg_accumulator_dtype=<class 'jax.numpy.int32'>,
                             local_aqt=None,
                             jax_scope_name='aqt_fwd',
                             allow_dummy_gradient_into_qtensor=False),
           dlhs=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=8,
                                                                                  preserve_zero=True,
                                                                                  preserve_max_val=False,
                                                                                  clip=True,
                                                                                  clip_gradient=False,
                                                                                  round=True,
                                                                                  noise_fn=JaxUniform(),
                                                                                  dtype=<class 'jax.numpy.int8'>),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=None,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(quantizer=Quantizer(numerics=IntNumerics(bits=8,
                                                                                  preserve_zero=True,
                                                                                  preserve_max_val=False,
                                                                                  clip=True,
                                                                                  clip_gradient=False,
                                                                                  round=True,
                                                                                  noise_fn=None,
                                                                                  dtype=<class 'jax.numpy.int8'>),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_accumulator_dtype=<class 'jax.numpy.int32'>,
                              local_aqt=None,
                              jax_scope_name='aqt_dlhs',
                              allow_dummy_gradient_into_qtensor=False),
           drhs=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=NoNumerics(noise_fn=JaxUniform(),
                                                                                 dtype=None),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=None,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(quantizer=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                 dtype=None),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_accumulator_dtype=None,
                              local_aqt=None,
                              jax_scope_name='aqt_drhs',
                              allow_dummy_gradient_into_qtensor=False))"""
    utils.test_pprint_eq(config.config_v4(), expected_cfg_str)

  def test_config_fwd_fp8(self):
    expected_cfg = """DotGeneral(fwd=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=Fp8Numerics(dtype=<class 'jax.numpy.float8_e4m3fn'>,
                                                                                 exponent_bits=4,
                                                                                 mantissa_bits=3,
                                                                                 noise_fn=None),
                                                            calib_shared_axes=None,
                                                            scale_stop_grad=True,
                                                            calibration=AbsMaxCalibration(scale=None),
                                                            po2_scale=False,
                                                            context=Context(key=None,
                                                                            train_step=None)),
                                        use_fwd_quant=None,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             rhs=Tensor(quantizer=Quantizer(numerics=Fp8Numerics(dtype=<class 'jax.numpy.float8_e4m3fn'>,
                                                                                 exponent_bits=4,
                                                                                 mantissa_bits=3,
                                                                                 noise_fn=None),
                                                            calib_shared_axes=None,
                                                            scale_stop_grad=True,
                                                            calibration=AbsMaxCalibration(scale=None),
                                                            po2_scale=False,
                                                            context=Context(key=None,
                                                                            train_step=None)),
                                        use_fwd_quant=None,
                                        dequant_mode=<DequantMode.OUTPUT: 1>,
                                        calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                             dg_accumulator_dtype=<class 'jax.numpy.float32'>,
                             local_aqt=None,
                             jax_scope_name='aqt_fwd',
                             allow_dummy_gradient_into_qtensor=False),
           dlhs=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                 dtype=None),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=None,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(quantizer=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                 dtype=None),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_accumulator_dtype=None,
                              local_aqt=None,
                              jax_scope_name='aqt_dlhs',
                              allow_dummy_gradient_into_qtensor=False),
           drhs=DotGeneralRaw(lhs=Tensor(quantizer=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                 dtype=None),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=None,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              rhs=Tensor(quantizer=Quantizer(numerics=NoNumerics(noise_fn=None,
                                                                                 dtype=None),
                                                             calib_shared_axes=None,
                                                             scale_stop_grad=True,
                                                             calibration=AbsMaxCalibration(scale=None),
                                                             po2_scale=False,
                                                             context=Context(key=None,
                                                                             train_step=None)),
                                         use_fwd_quant=False,
                                         dequant_mode=<DequantMode.OUTPUT: 1>,
                                         calibration_mode=<CalibrationMode.CONTRACTING_AXIS: 1>),
                              dg_accumulator_dtype=None,
                              local_aqt=None,
                              jax_scope_name='aqt_drhs',
                              allow_dummy_gradient_into_qtensor=False))"""
    utils.test_pprint_eq(config.config_fwd_fp8(), expected_cfg)

  def test_set_int_numerics_preserve_zero(self):
    cfg = config.config_v4()
    for tensor in [cfg.fwd.lhs, cfg.fwd.rhs, cfg.dlhs.lhs, cfg.dlhs.rhs]:
      self.assertTrue(tensor.quantizer.numerics.preserve_zero)
    for tensor in [cfg.fwd.lhs, cfg.fwd.rhs, cfg.dlhs.lhs, cfg.dlhs.rhs]:
      self.assertEqual(tensor.quantizer.numerics.dtype, jnp.int8)

    config.set_int_numerics_preserve_zero(cfg, preserve_zero=False)
    for tensor in [cfg.fwd.lhs, cfg.fwd.rhs, cfg.dlhs.lhs, cfg.dlhs.rhs]:
      self.assertFalse(tensor.quantizer.numerics.preserve_zero)
    for tensor in [cfg.fwd.lhs, cfg.fwd.rhs, cfg.dlhs.lhs, cfg.dlhs.rhs]:
      self.assertIsNone(tensor.quantizer.numerics.dtype)

  def test_set_absmax_calib_scale(self):
    cfg = config.config_v4()
    for tensor in [
        cfg.fwd.lhs,
        cfg.fwd.rhs,
        cfg.dlhs.lhs,
        cfg.dlhs.rhs,
        cfg.drhs.lhs,
        cfg.drhs.rhs,
    ]:
      self.assertIsNone(tensor.quantizer.calibration.scale)
    for tensor in [cfg.fwd.lhs, cfg.fwd.rhs, cfg.dlhs.lhs, cfg.dlhs.rhs]:
      self.assertFalse(tensor.quantizer.numerics.clip_gradient)

    config.set_absmax_calib_scale(cfg, scale=3)
    for tensor in [
        cfg.fwd.lhs,
        cfg.fwd.rhs,
        cfg.dlhs.lhs,
        cfg.dlhs.rhs,
        cfg.drhs.lhs,
        cfg.drhs.rhs,
    ]:
      self.assertAlmostEqual(tensor.quantizer.calibration.scale, 3)
    for tensor in [cfg.fwd.lhs, cfg.fwd.rhs, cfg.dlhs.lhs, cfg.dlhs.rhs]:
      self.assertFalse(tensor.quantizer.numerics.clip_gradient)

    config.set_absmax_calib_scale(cfg, scale=0.1)
    for tensor in [cfg.fwd.lhs, cfg.fwd.rhs, cfg.dlhs.lhs, cfg.dlhs.rhs]:
      self.assertTrue(tensor.quantizer.numerics.clip_gradient)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for imagenet.models."""

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized

from aqt.jax_legacy.jax import flax_layers as aqt_flax_layers  # pylint: disable=unused-import
from aqt.jax_legacy.jax import hlo_utils
from aqt.jax_legacy.jax.imagenet import hparams_config
from aqt.jax_legacy.jax.imagenet import models
from aqt.jax_legacy.jax.imagenet.configs import resnet101_bfloat16
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4_a4_fixed
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w8
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w8_a8_fixed
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_bfloat16
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_w4_a4_auto
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_w8_a8_auto
from aqt.jax_legacy.jax.imagenet.train_utils import create_model
from aqt.jax_legacy.utils import hparams_utils

from jax import random
import jax.numpy as jnp

FLAGS = flags.FLAGS


class ModelsTest(parameterized.TestCase):

  def setUp(self):
    super(ModelsTest, self).setUp()
    self.rng_key = random.PRNGKey(0)

  def _check_weight_and_act(self, weight_prec, quant_act):
    quant_ops = 0
    if weight_prec:
      quant_ops += 1
    if quant_act and quant_act.prec:
      quant_ops += 1
    return quant_ops

  def _num_dense_floors(self, hparams):
    # Each dense layer can have one weight quant op and one input quant op.
    dense_floors = 0
    # There is 1 dense layer in the ResNet model.
    dense_floors += self._check_weight_and_act(hparams.dense_layer.weight_prec,
                                               hparams.dense_layer.quant_act)

    return dense_floors

  def _num_conv_floors(self, hparams):
    # Each conv layer can have one weight quant op and one input quant op.
    conv_floors = 0
    # There is 1 conv layer at ResNet model input (conv_init).
    conv_floors += self._check_weight_and_act(hparams.conv_init.weight_prec,
                                              hparams.conv_init.quant_act)
    for block in hparams.residual_blocks:
      # There are 3 conv layers per ResNet block.
      conv_floors += self._check_weight_and_act(block.conv_1.weight_prec,
                                                block.conv_1.quant_act)
      conv_floors += self._check_weight_and_act(block.conv_2.weight_prec,
                                                block.conv_2.quant_act)
      conv_floors += self._check_weight_and_act(block.conv_3.weight_prec,
                                                block.conv_3.quant_act)
      # There is 1 projection conv layer per group of blocks.
      if block.conv_proj is not None:
        conv_floors += self._check_weight_and_act(block.conv_proj.weight_prec,
                                                  block.conv_proj.quant_act)

    return conv_floors

  @parameterized.named_parameters(
      # Unquantized
      dict(
          testcase_name='resnet50_quantization_none',
          base_config_filename=resnet50_bfloat16),
      # Weights only
      dict(
          testcase_name='resnet50_quantization_weights_only_8',
          base_config_filename=resnet50_w8),
      dict(
          testcase_name='resnet50_quantization_weights_only_4',
          base_config_filename=resnet50_w4),
      # Weights and activations (fixed)
      dict(
          testcase_name='resnet50_quantization_weights_8_fixed_acts_8',
          base_config_filename=resnet50_w8_a8_fixed),
      dict(
          testcase_name='resnet50_quantization_weights_4_fixed_acts_4',
          base_config_filename=resnet50_w4_a4_fixed),
      # Weights and activations (automatic)
      dict(
          testcase_name='resnet50_quantization_weights_8_auto_acts_8',
          base_config_filename=resnet50_w8_a8_auto),
      dict(
          testcase_name='resnet50_quantization_weights_4_auto_acts_4',
          base_config_filename=resnet50_w4_a4_auto),
  )  # pylint: disable=line-too-long
  def test_create_model_object(self, base_config_filename):
    hparams = hparams_utils.load_hparams_from_config_dict(
        hparams_config.TrainingHParams, models.ResNet.HParams,
        base_config_filename.get_config())
    model, _ = create_model(
        self.rng_key, 32, 16, jnp.float32, hparams.model_hparams, train=True)
    self.assertIsNotNone(model)

  @parameterized.named_parameters(
      # Unquantized
      dict(
          testcase_name='resnet50_quantization_none',
          expected_layers=50,
          expected_layer_stages=[3, 4, 6, 3],
          base_config_filename=resnet50_bfloat16,
      ),
      dict(
          testcase_name='resnet101_quantization_none',
          expected_layers=101,
          expected_layer_stages=[3, 4, 23, 3],
          base_config_filename=resnet101_bfloat16,
      ),
      # Weights only
      dict(
          testcase_name='resnet50_quantization_weights_only_8',
          expected_layers=50,
          expected_layer_stages=[3, 4, 6, 3],
          base_config_filename=resnet50_w8),
      # Weights and activations (fixed)
      dict(
          testcase_name='resnet50_quantization_weights_8_fixed_acts_8',
          expected_layers=50,
          expected_layer_stages=[3, 4, 6, 3],
          base_config_filename=resnet50_w8_a8_fixed),
      # Weights and activations (automatic)
      dict(
          testcase_name='resnet50_quantization_weights_8_auto_acts_8',
          expected_layers=50,
          expected_layer_stages=[3, 4, 6, 3],
          base_config_filename=resnet50_w8_a8_auto),
  )  # pylint: disable=line-too-long
  def test_count_resnet_layers(self, expected_layers, expected_layer_stages,
                               base_config_filename):
    counted_layers = 0
    layer_stages = []
    hparams = hparams_utils.load_hparams_from_config_dict(
        hparams_config.TrainingHParams, models.ResNet.HParams,
        base_config_filename.get_config())
    # Count the layers in ResNet hparams
    if hasattr(hparams.model_hparams, 'dense_layer'):
      counted_layers += 1
    if hasattr(hparams.model_hparams, 'conv_init'):
      counted_layers += 1
    if hasattr(hparams.model_hparams, 'residual_blocks'):
      stage_block_count = 0
      for block in hparams.model_hparams.residual_blocks:
        stage_block_count += 1
        if hasattr(block, 'conv_1'):
          counted_layers += 1
        if hasattr(block, 'conv_2'):
          counted_layers += 1
        if hasattr(block, 'conv_3'):
          counted_layers += 1
        if hasattr(block, 'conv_proj') and block.conv_proj is not None:
          layer_stages.append(stage_block_count)
          stage_block_count = 0
      layer_stages.append(stage_block_count + 1)

    # If the first layer in residual blocks is a projection layer,
    #   the first element in layer_stages represents no blocks.
    if layer_stages[0] == 1:
      layer_stages = layer_stages[1:]

    self.assertEqual(counted_layers, expected_layers)
    self.assertEqual(layer_stages, expected_layer_stages)

  @parameterized.named_parameters(
      # Unquantized
      dict(
          testcase_name='resnet50_quantization_none',
          expected_floor_count=0,
          base_config_filename=resnet50_bfloat16,
      ),
      # Weights only
      dict(
          testcase_name='resnet50_quantization_weights_only_8',
          expected_floor_count=54,
          base_config_filename=resnet50_w8),
      # Weights and activations (fixed)
      dict(
          testcase_name='resnet50_quantization_weights_8_fixed_acts_8',
          expected_floor_count=108,
          base_config_filename=resnet50_w8_a8_fixed),
      # # Weights and activations (automatic)
      # Disabled because of floor count mismatch after JAX change.
      # dict(
      #     testcase_name='resnet50_quantization_weights_8_auto_acts_8',
      #     expected_floor_count=108,
      #     base_config_filename=resnet50_w8_a8_auto),
  )  # pylint: disable=line-too-long
  def test_count_floor_ops(self, base_config_filename, expected_floor_count):
    hparams = hparams_utils.load_hparams_from_config_dict(
        hparams_config.TrainingHParams, models.ResNet.HParams,
        base_config_filename.get_config())
    input_shape = (32, 16, 16, 3)
    model, init_state = create_model(
        self.rng_key,
        input_shape[0],
        input_shape[1],
        jnp.float32,
        hparams.model_hparams,
        train=False)
    hlo_proto = hlo_utils.load_hlo_proto_from_model(model, init_state,
                                                    [input_shape])
    floor_count = hlo_utils.count_ops_in_hlo_proto(hlo_proto, r'floor')
    self.assertEqual(floor_count, expected_floor_count)
    # Expected floor count
    expected_floor_count_from_hparams = 0
    expected_floor_count_from_hparams += self._num_dense_floors(
        hparams.model_hparams)
    expected_floor_count_from_hparams += self._num_conv_floors(
        hparams.model_hparams)
    self.assertEqual(floor_count, expected_floor_count_from_hparams)


if __name__ == '__main__':
  FLAGS.metadata_enabled = True  # Passes quantization information to HLO
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ImageNet example with optional quantization.

This script trains a ResNet model on the ImageNet dataset.
The data is loaded using tensorflow_datasets.
The function create_model() can accept an hparam file as input, which will
  determine the model's size, training parameters, quantization precisions, etc.
"""

import functools
import json
import time

from absl import app
from absl import flags
from absl import logging

from aqt.jax_legacy.jax import compute_cost_utils
from aqt.jax_legacy.jax import hlo_utils
from aqt.jax_legacy.jax import train_utils
from aqt.jax_legacy.jax.imagenet import hparams_config
from aqt.jax_legacy.jax.imagenet import input_pipeline
from aqt.jax_legacy.jax.imagenet import models
from aqt.jax_legacy.jax.imagenet import train_utils as imagenet_train_utils
from aqt.jax_legacy.jax.imagenet.configs.paper.resnet50_w8_a8_auto import get_config as w8a8auto_paper_config
from aqt.jax_legacy.utils import hparams_utils as os_hparams_utils
from aqt.jax_legacy.utils import report_utils
from aqt.jax_legacy.utils import summary_utils

from flax import jax_utils
from flax.metrics import tensorboard
from flax.training import checkpoints
from flax.training import common_utils
from flax.training import dynamic_scale as dynamic_scale_lib
import jax
from jax import random
import jax.nn
import jax.numpy as jnp
from ml_collections import config_flags
import optax
import tensorflow.compat.v2 as tf
from tensorflow.io import gfile


COMPUTE_MEMORY_COST_FILENAME = 'compute_memory_cost.json'

FLAGS = flags.FLAGS

flags.DEFINE_string(
    'model_dir', default=None, help=('Directory to store model data.'))

flags.DEFINE_string(
    'data_dir', default=None, help=('Directory where imagenet data is stored.'))

config_flags.DEFINE_config_file('hparams_config_dict', None,
                                'Path to file defining a config dict.')

flags.DEFINE_integer(
    'config_idx',
    default=None,
    help=(
        'Identifies which config within the sweep this training run should use.'
    ))

flags.DEFINE_integer(
    'batch_size', default=128, help=('Batch size for training.'))

flags.DEFINE_bool('cache', default=False, help=('If True, cache the dataset.'))

flags.DEFINE_string(
    'resnet508b_ckpt_path',
    default=None,
    help=('Path to the checkpoint directory of ResNet508b as a teacher model.'))

flags.DEFINE_string(
    'report_dir',
    default=None,
    help=('Directory to save experiment report to after training has '
          'completed. '))

flags.DEFINE_bool(
    'half_precision',
    default=True,
    help=('If bfloat16/float16 should be used instead of float32.'))

flags.DEFINE_integer(
    'state_dict_summary_freq',
    default=200,
    help='Number of training steps between state dict summaries reported to '
    'Tensorboard. Relevant to --visualize_acts_bound and --collect_acts_stats.')

flags.DEFINE_float(
    'step_lr_coeff',
    default=0.2,
    help='Coefficient used by the step lr scheduler to decay at the end of'
    ' every interval.')

flags.DEFINE_integer(
    'step_lr_intervals',
    default=6,
    help='Number of intervals in the step lr scheduler')

flags.DEFINE_bool(
    'visualize_acts_bound',
    default=False,
    help=(
        'Whether to visualize activations bounds for auto-clip in Tensorboard.'
        ' The bounds appear as scalar and will be named as "GetBounds_0/bounds"'
        ' prefixed with the all the parents module name.'))

flags.DEFINE_bool(
    'estimate_compute_and_memory_cost',
    default=False,
    help='Whether to estimate compute and memory cost.')

flags.DEFINE_bool(
    'write_summary',
    default=False,
    help='Whether to write the summary_dict to tensorboard.')


def cosine_decay(base_learning_rate, step, decay_steps, alpha=0.001):
  ratio = jnp.minimum(jnp.maximum(0., step / decay_steps), 1.)
  decay = 0.5 * (1. + jnp.cos(jnp.pi * ratio))
  decayed = (1 - alpha) * decay + alpha
  return decayed * base_learning_rate


def step_decay(lr, step, interval):
  """Constant LR within every interval, exponential decay at the end of each."""
  decays = (step // interval) - (step >= interval) + (step >= 2 * interval)
  lr *= FLAGS.step_lr_coeff**decays
  return lr


def create_learning_rate_fn(base_learning_rate, steps_per_epoch, lr_scheduler,
                            batch_size):
  """Create learning rate scheduler function."""
  num_epochs = lr_scheduler.num_epochs
  warmup_steps = lr_scheduler.warmup_epochs * steps_per_epoch
  cooldown_steps = lr_scheduler.cooldown_epochs * steps_per_epoch
  total_steps = num_epochs * steps_per_epoch

  def step_fn(step):
    if lr_scheduler.scheduler == 'linear':
      epochs_left = num_epochs - step / steps_per_epoch
      lr = epochs_left / num_epochs * base_learning_rate
    elif lr_scheduler.scheduler == 'piecewise':
      knee_steps = lr_scheduler.knee_epochs * steps_per_epoch
      knee_lr = lr_scheduler.knee_lr * batch_size / 256.
      endlr = lr_scheduler.endlr * batch_size / 256.
      # compute phase2 lr
      k1 = step / knee_steps
      lr1 = k1 * knee_lr + (1 - k1) * base_learning_rate
      k2 = (step - knee_steps) / (total_steps - knee_steps)
      lr2 = k2 * endlr + (1 - k2) * knee_lr
      lr = jnp.where(step >= knee_steps, lr2, lr1)
    elif lr_scheduler.scheduler == 'cosine':
      lr = cosine_decay(base_learning_rate, step - warmup_steps,
                        total_steps - warmup_steps - cooldown_steps)
    elif lr_scheduler.scheduler == 'step':
      epoch = step / steps_per_epoch
      lr = step_decay(base_learning_rate, epoch,
                      num_epochs // FLAGS.step_lr_intervals)
    else:
      raise ValueError('Invalid learning rate scheduler.')
    if warmup_steps > 0:
      warmup = jnp.minimum(1., step / warmup_steps)
      lr *= warmup
    if cooldown_steps > 0:
      cooldown = jnp.minimum(1., (total_steps - step) / cooldown_steps)
      lr *= cooldown
    return lr

  return step_fn


# flax.struct.dataclass enables instances of this class to be passed into jax
# transformations like tree_map and pmap.
def estimate_compute_and_memory_cost(image_size, model_dir, hparams) -> None:
  """Estimate compute and memory cost of model."""
  FLAGS.metadata_enabled_temp = True
  input_shape = (1, image_size, image_size, 3)
  model, init_state = imagenet_train_utils.create_model(
      jax.random.PRNGKey(0),
      input_shape[0],
      input_shape[1],
      jnp.float32,
      hparams.model_hparams,
      train=False)
  hlo_proto = hlo_utils.load_hlo_proto_from_model(model, init_state,
                                                  [input_shape])
  del model, init_state
  cost_dict = compute_cost_utils.estimate_compute_cost(hlo_proto)
  memory_cost_dict = compute_cost_utils.estimate_memory_cost(hlo_proto)
  cost_dict.update(memory_cost_dict)
  FLAGS.metadata_enabled_temp = False

  with gfile.GFile(gfile.join(model_dir, COMPUTE_MEMORY_COST_FILENAME),
                   'w') as file:
    json.dump(cost_dict, file, indent=2)
  logging.info('Estimated compute and memory costs and wrote to file')
  return


def restore_checkpoint(state):
  return checkpoints.restore_checkpoint(FLAGS.model_dir, state)


def save_checkpoint(state):
  if jax.host_id() == 0:
    # get train state from the first replica
    state = jax.device_get(jax.tree_map(lambda x: x[0], state))
    step = int(state.step)
    checkpoints.save_checkpoint(FLAGS.model_dir, state, step, keep=3)


def _get_state_dict_keys_from_flags():
  """Returns key suffixes to look up in flax state dict."""
  state_dict_keys = []
  if FLAGS.visualize_acts_bound:
    state_dict_keys.append('bounds')
  return state_dict_keys


def prepare_and_save_report(
    hparams: hparams_config.TrainingHParams,
    eval_freq: int,
    num_train_steps: int) -> None:
  """Creates a report for this run and saved it to FLAGS.report_dir."""
  # For models with quantization, we should only do early stopping after the
  # model is quantized during training, so after
  # hparams.activation_bound_start_step. (For models with quantized weights,
  # weights are quantized from the beginning of training.)
  # However, for unquantized activations, hparams.activation_bound_start_step is
  # set to -1, but the report start_step cannot be negative, so the max()
  # sets the start_step to 0 in this case.
  report_start_step = max(0, hparams.activation_bound_start_step)
  create_end_of_training_report = report_utils.create_end_of_training_report
  report = create_end_of_training_report(
      model_dir=FLAGS.model_dir,
      eval_freq=eval_freq,
      num_train_steps=num_train_steps,
      early_stop_attr='eval_accuracy',
      early_stop_agg=report_utils.MinOrMax.MAX,
      early_stop_ds_dir=None,
      smoothing_kernel=report_utils.SmoothingKernel.TRIANGULAR,
      other_ds_dirs=None,
      tags_to_include=[
          'eval_accuracy', 'eval_loss', 'train_accuracy', 'train_loss',
          'train_learning_rate'
      ],
      window_size_in_steps=eval_freq * 5,
      start_step=report_start_step,
  )
  logging.info('report created.')
  if FLAGS.report_dir is None:
    report_dir = FLAGS.model_dir
  else:
    report_dir = FLAGS.report_dir
  report_path = report_utils.save_report(report, report_dir)
  logging.info('Report saved to %s.', str(report_path))


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  tf.enable_v2_behavior()
  # make sure tf does not allocate gpu memory
  tf.config.experimental.set_visible_devices([], 'GPU')

  if jax.host_id() == 0:
    summary_writer = tensorboard.SummaryWriter(FLAGS.model_dir)

  image_size = 224

  batch_size = FLAGS.batch_size
  if batch_size % jax.device_count() > 0:
    raise ValueError('Batch size must be divisible by the number of devices')
  local_batch_size = batch_size // jax.host_count()
  device_batch_size = batch_size // jax.device_count()

  platform = jax.local_devices()[0].platform

  dynamic_scale = None
  if FLAGS.half_precision:
    if platform == 'tpu':
      model_dtype = jnp.bfloat16
      input_dtype = tf.bfloat16
    else:
      model_dtype = jnp.float16
      input_dtype = tf.float16
      dynamic_scale = dynamic_scale_lib.DynamicScale()
  else:
    model_dtype = jnp.float32
    input_dtype = tf.float32

  train_iter = imagenet_train_utils.create_input_iter(
      local_batch_size,
      FLAGS.data_dir,
      image_size,
      input_dtype,
      train=True,
      cache=FLAGS.cache)
  eval_iter = imagenet_train_utils.create_input_iter(
      local_batch_size,
      FLAGS.data_dir,
      image_size,
      input_dtype,
      train=False,
      cache=FLAGS.cache)

  # Create the hyperparameter object
  if FLAGS.hparams_config_dict:
    # In this case, there are multiple training configs defined in the config
    # dict, so we pull out the one this training run should use.
    if 'configs' in FLAGS.hparams_config_dict:
      hparams_config_dict = FLAGS.hparams_config_dict.configs[FLAGS.config_idx]
    else:
      hparams_config_dict = FLAGS.hparams_config_dict
    hparams = os_hparams_utils.load_hparams_from_config_dict(
        hparams_config.TrainingHParams, models.ResNet.HParams,
        hparams_config_dict)
  else:
    raise ValueError('Please provide a base config dict.')

  os_hparams_utils.write_hparams_to_file_with_host_id_check(
      hparams, FLAGS.model_dir)

  # get num_epochs from hparam instead of FLAGS
  num_epochs = hparams.lr_scheduler.num_epochs
  steps_per_epoch = input_pipeline.TRAIN_IMAGES // batch_size
  steps_per_eval = input_pipeline.EVAL_IMAGES // batch_size
  steps_per_checkpoint = steps_per_epoch * 10
  num_steps = steps_per_epoch * num_epochs

  # Estimate compute / memory costs
  if jax.host_id() == 0 and FLAGS.estimate_compute_and_memory_cost:
    estimate_compute_and_memory_cost(
        image_size=image_size, model_dir=FLAGS.model_dir, hparams=hparams)
    logging.info('Writing training HLO and estimating compute/memory costs.')

  rng = random.PRNGKey(hparams.seed)
  model, variables = imagenet_train_utils.create_model(
      rng,
      device_batch_size,
      image_size,
      model_dtype,
      hparams=hparams.model_hparams,
      train=True,
      is_teacher=hparams.is_teacher)

  # pylint: disable=g-long-lambda
  if hparams.teacher_model == 'resnet50-8bit':
    teacher_config = w8a8auto_paper_config()
    teacher_hparams = os_hparams_utils.load_hparams_from_config_dict(
        hparams_config.TrainingHParams, models.ResNet.HParams, teacher_config)
    teacher_model, _ = imagenet_train_utils.create_model(
        rng,
        device_batch_size,
        image_size,
        model_dtype,
        hparams=teacher_hparams.model_hparams,
        train=False,
        is_teacher=True)  # teacher model does not need to be trainable
    # Directory where checkpoints are saved
    ckpt_model_dir = FLAGS.resnet508b_ckpt_path
    # will restore to best checkpoint
    state_load = checkpoints.restore_checkpoint(ckpt_model_dir, None)
    teacher_variables = {'params': state_load['optimizer']['target']}
    teacher_variables.update(state_load['model_state'])
    # create a dictionary for better argument passing
    teacher = {
        'model':
            lambda var, img, labels: jax.nn.softmax(
                teacher_model.apply(var, img)),
        'variables':
            teacher_variables,
    }
  elif hparams.teacher_model == 'labels':
    teacher = {
        'model':
            lambda var, img, labels: common_utils.onehot(
                labels, num_classes=1000),
        'variables': {},  # no need of variables in this case
    }
  else:
    raise ValueError('The specified teacher model is not supported.')

  mutable_vars, params = variables.pop('params')
  base_learning_rate = hparams.base_learning_rate * batch_size / 256.
  learning_rate_fn = create_learning_rate_fn(base_learning_rate,
                                             steps_per_epoch,
                                             hparams.lr_scheduler,
                                             batch_size)
  if hparams.optimizer == 'sgd':
    tx = optax.sgd(learning_rate=learning_rate_fn,
                   momentum=hparams.momentum,
                   nesterov=True)
  elif hparams.optimizer == 'adam':
    tx = optax.adam(learning_rate=learning_rate_fn,
                    b1=hparams.adam.beta1,
                    b2=hparams.adam.beta2)
  else:
    raise ValueError('Optimizer type is not supported.')
  state = imagenet_train_utils.TrainState.create(
      params=params,
      tx=tx,
      mutable_vars=mutable_vars,
      dynamic_scale=dynamic_scale)

  state = restore_checkpoint(state)
  step_offset = int(state.step)  # step_offset > 0 if restarting from checkpoint
  state = jax_utils.replicate(state)

  p_train_step = jax.pmap(
      functools.partial(
          imagenet_train_utils.train_step,
          model,
          learning_rate_fn=learning_rate_fn,
          teacher=teacher),
      axis_name='batch',
      static_broadcasted_argnums=(2, 3, 4))
  p_eval_step = jax.pmap(
      functools.partial(imagenet_train_utils.eval_step, model),
      axis_name='batch',
      static_broadcasted_argnums=(2,))

  epoch_metrics = []
  state_dict_summary_all = []
  state_dict_keys = _get_state_dict_keys_from_flags()
  t_loop_start = time.time()
  last_log_step = 0
  for step, batch in zip(range(step_offset, num_steps), train_iter):
    if hparams.early_stop_steps >= 0 and step > hparams.early_stop_steps * steps_per_epoch:
      break
    update_bounds = train_utils.should_update_bounds(
        hparams.activation_bound_update_freq,
        hparams.activation_bound_start_step, step)
    # and pass the result bool value to p_train_step
    # The function should take hparams.weight_quant_start_step as inputs
    quantize_weights = train_utils.should_quantize_weights(
        hparams.weight_quant_start_step, step // steps_per_epoch)
    state, metrics = p_train_step(state, batch, hparams, update_bounds,
                                  quantize_weights)

    state_dict_summary = summary_utils.get_state_dict_summary(
        state.mutable_vars, state_dict_keys)
    state_dict_summary_all.append(state_dict_summary)

    epoch_metrics.append(metrics)
    def should_log(step):
      epoch_no = step // steps_per_epoch
      step_in_epoch = step - epoch_no * steps_per_epoch
      do_log = False
      do_log = do_log or (step + 1 == num_steps)  # log at the end
      end_of_train = step / num_steps > 0.9
      do_log = do_log or ((step_in_epoch %
                           (steps_per_epoch // 4) == 0) and not end_of_train)
      do_log = do_log or ((step_in_epoch %
                           (steps_per_epoch // 16) == 0) and end_of_train)
      return do_log

    if should_log(step):
      epoch = step // steps_per_epoch
      epoch_metrics = common_utils.get_metrics(epoch_metrics)
      summary = jax.tree_map(lambda x: x.mean(), epoch_metrics)
      logging.info('train epoch: %d, loss: %.4f, accuracy: %.2f', epoch,
                   summary['loss'], summary['accuracy'] * 100)
      steps_per_sec = (step - last_log_step) / (time.time() - t_loop_start)
      last_log_step = step
      t_loop_start = time.time()

      # Write to TensorBoard
      state_dict_summary_all = common_utils.get_metrics(state_dict_summary_all)
      if jax.host_id() == 0:
        for key, vals in epoch_metrics.items():
          tag = 'train_%s' % key
          for i, val in enumerate(vals):
            summary_writer.scalar(tag, val, step - len(vals) + i + 1)
        summary_writer.scalar('steps per second', steps_per_sec, step)

        if FLAGS.write_summary:
          summary_utils.write_state_dict_summaries_to_tb(
              state_dict_summary_all, summary_writer,
              FLAGS.state_dict_summary_freq, step)

      state_dict_summary_all = []
      epoch_metrics = []
      eval_metrics = []

      # sync batch statistics across replicas
      state = imagenet_train_utils.sync_batch_stats(state)
      for _ in range(steps_per_eval):
        eval_batch = next(eval_iter)
        metrics = p_eval_step(state, eval_batch, quantize_weights)
        eval_metrics.append(metrics)
      eval_metrics = common_utils.get_metrics(eval_metrics)
      summary = jax.tree_map(lambda x: x.mean(), eval_metrics)
      logging.info('eval epoch: %d, loss: %.4f, accuracy: %.2f', epoch,
                   summary['loss'], summary['accuracy'] * 100)
      if jax.host_id() == 0:
        for key, val in eval_metrics.items():
          tag = 'eval_%s' % key
          summary_writer.scalar(tag, val.mean(), step)
        summary_writer.flush()
    if (step + 1) % steps_per_checkpoint == 0 or step + 1 == num_steps:
      # Only save the report when step > activation_bound_start_step
      if jax.host_id() == 0 and step > hparams.activation_bound_start_step:
        prepare_and_save_report(
            hparams, eval_freq=steps_per_epoch, num_train_steps=step + 1)
      state = imagenet_train_utils.sync_batch_stats(state)
      save_checkpoint(state)

  # Wait until computations are done before exiting
  jax.random.normal(jax.random.PRNGKey(0), ()).block_until_ready()


if __name__ == '__main__':
  app.run(main)
